class DatasetContainsTargetError(ValueError):
    """
    Raised when a dataset contains the target column already.

    Parameters
    ----------
    target_name: str
        The name of the target column.
    """

    def __init__(self, target_name: str):
        super().__init__(f"Dataset already contains the target column '{target_name}'.")


class DatasetMissesFeaturesError(ValueError):
    """
    Raised when a dataset misses feature columns.

    Parameters
    ----------
    missing_feature_names: list[str]
        The names of the missing feature columns.
    """

    def __init__(self, missing_feature_names: list[str]):
        super().__init__(f"Dataset misses the feature columns '{missing_feature_names}'.")


class DatasetMissesDataError(ValueError):
    """Raised when a dataset contains no rows."""

    def __init__(self) -> None:
        super().__init__("Dataset contains no rows")


class LearningError(Exception):
    """
    Raised when an error occurred while training a model.

    Parameters
    ----------
    reason: str
        The reason for the error.
    """

    def __init__(self, reason: str):
        super().__init__(f"Error occurred while learning: {reason}")


class ModelNotFittedError(Exception):
    """Raised when a model is used before fitting it."""

    def __init__(self) -> None:
        super().__init__("The model has not been fitted yet.")


class PredictionError(Exception):
    """
    Raised when an error occurred while prediction a target vector using a model.

    Parameters
    ----------
    reason: str
        The reason for the error.
    """

    def __init__(self, reason: str):
        super().__init__(f"Error occurred while predicting: {reason}")


class UntaggedTableError(Exception):
    """Raised when an untagged table is used instead of a TaggedTable in a regression or classification."""

    def __init__(self) -> None:
        super().__init__(
            (
                "This method needs a tagged table.\nA tagged table is a table that additionally knows which columns are"
                " features and which are the target to predict.\nUse Table.tag_column() to create a tagged table."
            ),
        )



from __future__ import annotations

import copy
import functools
from collections.abc import Callable, Mapping
from typing import TYPE_CHECKING, Any

import pandas as pd

from safeds.data.tabular.typing import ColumnType, Schema
from safeds.exceptions import UnknownColumnNameError

if TYPE_CHECKING:
    from collections.abc import Iterator


class Row(Mapping[str, Any]):
    """
    A row is a collection of named values.

    Parameters
    ----------
    data : Mapping[str, Any] | None
        The data. If None, an empty row is created.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Row
    >>> row = Row({"a": 1, "b": 2})
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Creation
    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Row:
        """
        Create a row from a dictionary that maps column names to column values.

        Parameters
        ----------
        data : dict[str, Any]
            The data.

        Returns
        -------
        row : Row
            The created row.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row.from_dict({"a": 1, "b": 2})
        """
        return Row(data)

    @staticmethod
    def _from_pandas_dataframe(data: pd.DataFrame, schema: Schema | None = None) -> Row:
        """
        Create a row from a `pandas.DataFrame`.

        Parameters
        ----------
        data : pd.DataFrame
            The data.
        schema : Schema | None
            The schema. If None, the schema is inferred from the data.

        Returns
        -------
        row : Row
            The created row.

        Raises
        ------
        ValueError
            If the dataframe does not contain exactly one row.

        Examples
        --------
        >>> import pandas as pd
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row._from_pandas_dataframe(pd.DataFrame({"a": [1], "b": [2]}))
        """
        if data.shape[0] != 1:
            raise ValueError("The dataframe has to contain exactly one row.")

        data = data.reset_index(drop=True)

        result = object.__new__(Row)
        result._data = data

        if schema is None:
            # noinspection PyProtectedMember
            result._schema = Schema._from_pandas_dataframe(data)
        else:
            result._schema = schema

        return result

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, data: Mapping[str, Any] | None = None) -> None:
        """
        Create a row from a mapping of column names to column values.

        Parameters
        ----------
        data : Mapping[str, Any] | None
            The data. If None, an empty row is created.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        """
        if data is None:
            data = {}

        data = {key: [value] for key, value in data.items()}

        self._data: pd.DataFrame = pd.DataFrame(data)
        # noinspection PyProtectedMember
        self._schema: Schema = Schema._from_pandas_dataframe(self._data)

    def __contains__(self, obj: Any) -> bool:
        """
        Check whether the row contains an object as key.

        Parameters
        ----------
        obj : Any
            The object.

        Returns
        -------
        has_column : bool
            True, if the row contains the object as key, False otherwise.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> "a" in row
        True

        >>> "c" in row
        False
        """
        return isinstance(obj, str) and self.has_column(obj)

    def __eq__(self, other: Any) -> bool:
        """
        Check whether this row is equal to another object.

        Parameters
        ----------
        other : Any
            The other object.

        Returns
        -------
        equal : bool
            True if the other object is an identical row. False if the other object is a different row. NotImplemented
            if the other object is not a row.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row1 = Row({"a": 1, "b": 2})
        >>> row2 = Row({"a": 1, "b": 2})
        >>> row1 == row2
        True

        >>> row3 = Row({"a": 1, "b": 3})
        >>> row1 == row3
        False
        """
        if not isinstance(other, Row):
            return NotImplemented
        if self is other:
            return True
        return self._schema == other._schema and self._data.equals(other._data)

    def __getitem__(self, column_name: str) -> Any:
        """
        Return the value of a specified column.

        Parameters
        ----------
        column_name : str
            The column name.

        Returns
        -------
        value : Any
            The column value.

        Raises
        ------
        UnknownColumnNameError
            If the row does not contain the specified column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> row["a"]
        1
        """
        return self.get_value(column_name)

    def __iter__(self) -> Iterator[Any]:
        """
        Create an iterator for the column names of this row.

        Returns
        -------
        iterator : Iterator[Any]
            The iterator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> list(row)
        ['a', 'b']
        """
        return iter(self.column_names)

    def __len__(self) -> int:
        """
        Return the number of columns in this row.

        Returns
        -------
        number_of_columns : int
            The number of columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> len(row)
        2
        """
        return self._data.shape[1]

    def __repr__(self) -> str:
        """
        Return an unambiguous string representation of this row.

        Returns
        -------
        representation : str
            The string representation.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1})
        >>> repr(row)
        "Row({'a': 1})"
        """
        return f"Row({self!s})"

    def __str__(self) -> str:
        """
        Return a user-friendly string representation of this row.

        Returns
        -------
        representation : str
            The string representation.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1})
        >>> str(row)
        "{'a': 1}"
        """
        match len(self):
            case 0:
                return "{}"
            case 1:
                return str(self.to_dict())
            case _:
                lines = (f"    {name!r}: {value!r}" for name, value in self.to_dict().items())
                joined = ",\n".join(lines)
                return f"{{\n{joined}\n}}"

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def column_names(self) -> list[str]:
        """
        Return a list of all column names in the row.

        Returns
        -------
        column_names : list[str]
            The column names.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> row.column_names
        ['a', 'b']
        """
        return self._schema.column_names

    @property
    def number_of_column(self) -> int:
        """
        Return the number of columns in this row.

        Returns
        -------
        number_of_column : int
            The number of columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> row.number_of_column
        2
        """
        return self._data.shape[1]

    @property
    def schema(self) -> Schema:
        """
        Return the schema of the row.

        Returns
        -------
        schema : Schema
            The schema.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> schema = row.schema
        """
        return self._schema

    # ------------------------------------------------------------------------------------------------------------------
    # Getters
    # ------------------------------------------------------------------------------------------------------------------

    def get_value(self, column_name: str) -> Any:
        """
        Return the value of a specified column.

        Parameters
        ----------
        column_name : str
            The column name.

        Returns
        -------
        value : Any
            The column value.

        Raises
        ------
        UnknownColumnNameError
            If the row does not contain the specified column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> row.get_value("a")
        1
        """
        if not self.has_column(column_name):
            raise UnknownColumnNameError([column_name])

        return self._data.loc[0, column_name]

    def has_column(self, column_name: str) -> bool:
        """
        Check whether the row contains a given column.

        Parameters
        ----------
        column_name : str
            The column name.

        Returns
        -------
        has_column : bool
            True, if the row contains the column, False otherwise.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> row.has_column("a")
        True

        >>> row.has_column("c")
        False
        """
        return self._schema.has_column(column_name)

    def get_column_type(self, column_name: str) -> ColumnType:
        """
        Return the type of the specified column.

        Parameters
        ----------
        column_name : str
            The column name.

        Returns
        -------
        type : ColumnType
            The type of the column.

        Raises
        ------
        UnknownColumnNameError
            If the row does not contain the specified column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> row.get_column_type("a")
        Integer
        """
        return self._schema.get_column_type(column_name)

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    def sort_columns(
        self,
        comparator: Callable[[tuple, tuple], int] = lambda col1, col2: (col1[0] > col2[0]) - (col1[0] < col2[0]),
    ) -> Row:
        """
        Sort the columns of a `Row` with the given comparator and return a new `Row`.

        The original row is not modified. The comparator is a function that takes two tuples of (ColumnName,
        Value) `col1` and `col2` and returns an integer:

        * If `col1` should be ordered before `col2`, the function should return a negative number.
        * If `col1` should be ordered after `col2`, the function should return a positive number.
        * If the original order of `col1` and `col2` should be kept, the function should return 0.

        If no comparator is given, the columns will be sorted alphabetically by their name.

        Parameters
        ----------
        comparator : Callable[[tuple, tuple], int]
            The function used to compare two tuples of (ColumnName, Value).

        Returns
        -------
        new_row : Row
            A new row with sorted columns.
        """
        sorted_row_dict = dict(sorted(self.to_dict().items(), key=functools.cmp_to_key(comparator)))
        return Row.from_dict(sorted_row_dict)

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    def to_dict(self) -> dict[str, Any]:
        """
        Return a dictionary that maps column names to column values.

        Returns
        -------
        data : dict[str, Any]
            Dictionary representation of the row.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> row.to_dict()
        {'a': 1, 'b': 2}
        """
        return {column_name: self.get_value(column_name) for column_name in self.column_names}

    def to_html(self) -> str:
        """
        Return an HTML representation of the row.

        Returns
        -------
        output : str
            The generated HTML.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row
        >>> row = Row({"a": 1, "b": 2})
        >>> html = row.to_html()
        """
        return self._data.to_html(max_rows=1, max_cols=self._data.shape[1])

    # ------------------------------------------------------------------------------------------------------------------
    # IPython integration
    # ------------------------------------------------------------------------------------------------------------------

    def _repr_html_(self) -> str:
        """
        Return an HTML representation of the row.

        Returns
        -------
        output : str
            The generated HTML.
        """
        return self._data.to_html(max_rows=1, max_cols=self._data.shape[1], notebook=True)

    # ------------------------------------------------------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------------------------------------------------------

    def _copy(self) -> Row:
        """
        Return a copy of this row.

        Returns
        -------
        copy : Row
            The copy of this row.
        """
        return copy.deepcopy(self)



from __future__ import annotations

from abc import ABC, abstractmethod

from numpy import isinf, isnan


class OutOfBoundsError(ValueError):
    """
    A generic exception that can be used to signal that a (float) value is outside its expected range.

    Parameters
    ----------
    actual: float
        The actual value that is outside its expected range.
    name: str | None
        The name of the offending variable.
    lower_bound: Bound | None
        The lower bound of the expected range. Use None if there is no lower Bound.
    upper_bound: Bound | None
        The upper bound of the expected range. Use None if there is no upper Bound.
    """

    def __init__(
        self,
        actual: float,
        *,
        name: str | None = None,
        lower_bound: Bound | None = None,
        upper_bound: Bound | None = None,
    ):
        """
        Initialize an OutOfBoundsError.

        Parameters
        ----------
        actual: float
            The actual value that is outside its expected range.
        name: str | None
            The name of the offending variable.
        lower_bound: Bound | None
            The lower bound of the expected range. Use None if there is no lower Bound.
        upper_bound: Bound | None
            The upper bound of the expected range. Use None if there is no upper Bound.

        Raises
        ------
        ValueError
            * If one of the given Bounds is +/-inf. (For infinite Bounds, pass None instead.)
            * If one of the given Bounds is nan.
            * If upper_bound < lower_bound.
            * If actual does not lie outside the given interval.
            * If actual is not a real number.
        """
        # Validate bound parameters:
        if lower_bound is None and upper_bound is None:
            raise ValueError("Illegal interval: Attempting to raise OutOfBoundsError, but no bounds given.")
        if (lower_bound is not None and isinf(lower_bound.value)) or (
            upper_bound is not None and isinf(upper_bound.value)
        ):
            raise ValueError("Illegal interval: Lower and upper bounds must be real numbers, or None if unbounded.")
        # Validate actual parameter:
        if isinf(actual) or isnan(actual):
            raise ValueError("Attempting to raise OutOfBoundsError with actual value not being a real number.")
        # Use local variables with stricter types to help static analysis:
        _lower_bound: Bound = lower_bound if lower_bound is not None else OpenBound(float("-inf"))
        _upper_bound: Bound = upper_bound if upper_bound is not None else OpenBound(float("inf"))
        # Check bounds:
        if _upper_bound.value < _lower_bound.value:
            raise ValueError(
                (
                    f"Illegal interval: Attempting to raise OutOfBoundsError, but given upper bound {_upper_bound} is "
                    f"actually less than given lower bound {_lower_bound}."
                ),
            )
        # Check that actual is indeed outside the interval:
        elif _lower_bound._check_lower_bound(actual) and _upper_bound._check_upper_bound(actual):
            raise ValueError(
                (
                    f"Illegal interval: Attempting to raise OutOfBoundsError, but value {actual} is not actually"
                    f" outside given interval {_lower_bound._str_lower_bound()}, {_upper_bound._str_upper_bound()}."
                ),
            )
        # Raise the actual exception:
        full_variable_name = actual if name is None else f"{name} (={actual})"
        super().__init__(
            f"{full_variable_name} is not inside {_lower_bound._str_lower_bound()}, {_upper_bound._str_upper_bound()}.",
        )


class Bound(ABC):
    """
    Abstract base class for (lower or upper) Bounds on a float value.

    Parameters
    ----------
    value: float
        The value of the Bound.
    """

    def __init__(self, value: float):
        """
        Initialize a Bound.

        Parameters
        ----------
        value: float
            The value of the Bound.

        Raises
        ------
        ValueError
            If value is nan or if value is +/-inf and the Bound type does not allow for infinite Bounds.
        """
        if isnan(value):
            raise ValueError("Bound must be a real number, not nan.")
        self._value = value

    def __str__(self) -> str:
        """Get a string representation of the concrete value of the Bound."""
        return str(self.value)

    @property
    def value(self) -> float:
        """Get the concrete value of the Bound."""
        return self._value

    @abstractmethod
    def _str_lower_bound(self) -> str:
        """Get a string representation of the Bound as the lower Bound of an interval."""

    @abstractmethod
    def _str_upper_bound(self) -> str:
        """Get a string representation of the Bound as the upper Bound of an interval."""

    @abstractmethod
    def _check_lower_bound(self, actual: float) -> bool:
        """
        Check that a value does not exceed the Bound on the lower side.

        Parameters
        ----------
        actual: float
            The actual value that should be checked for not exceeding the Bound.
        """

    @abstractmethod
    def _check_upper_bound(self, actual: float) -> bool:
        """
        Check that a value does not exceed the Bound on the upper side.

        Parameters
        ----------
        actual: float
            The actual value that should be checked for not exceeding the Bound.
        """


class ClosedBound(Bound):
    """
    A closed Bound, i.e. the value on the border belongs to the range.

    Parameters
    ----------
    value: float
        The value of the Bound.
    """

    def __init__(self, value: float):
        """
        Initialize a ClosedBound.

        Parameters
        ----------
        value: float
            The value of the ClosedBound.

        Raises
        ------
        ValueError
            If value is nan or if value is +/-inf.
        """
        if value == float("-inf") or value == float("inf"):
            raise ValueError("ClosedBound must be a real number, not +/-inf.")
        super().__init__(value)

    def _str_lower_bound(self) -> str:
        """Get a string representation of the ClosedBound as the lower Bound of an interval."""
        return f"[{self}"

    def _str_upper_bound(self) -> str:
        """Get a string representation of the ClosedBound as the upper Bound of an interval."""
        return f"{self}]"

    def _check_lower_bound(self, actual: float) -> bool:
        """
        Check that a value is not strictly lower than the ClosedBound.

        Parameters
        ----------
        actual: float
            The actual value that should be checked for not exceeding the Bound.
        """
        return actual >= self.value

    def _check_upper_bound(self, actual: float) -> bool:
        """
        Check that a value is not strictly higher than the ClosedBound.

        Parameters
        ----------
        actual: float
            The actual value that should be checked for not exceeding the Bound.
        """
        return actual <= self.value


class OpenBound(Bound):
    """
    An open Bound, i.e. the value on the border does not belong to the range.

    Parameters
    ----------
    value: float
        The value of the OpenBound.
    """

    def __init__(self, value: float):
        """
        Initialize an OpenBound.

        Parameters
        ----------
        value: float
            The value of the OpenBound.

        Raises
        ------
        ValueError
            If value is nan.
        """
        super().__init__(value)

    def __str__(self) -> str:
        """Get a string representation of the concrete value of the OpenBound."""
        if self.value == float("-inf"):
            return "-\u221e"
        elif self.value == float("inf"):
            return "\u221e"
        else:
            return super().__str__()

    def _str_lower_bound(self) -> str:
        """Get a string representation of the OpenBound as the lower Bound of an interval."""
        return f"({self}"

    def _str_upper_bound(self) -> str:
        """Get a string representation of the OpenBound as the upper Bound of an interval."""
        return f"{self})"

    def _check_lower_bound(self, actual: float) -> bool:
        """
        Check that a value is not lower or equal to the OpenBound.

        Parameters
        ----------
        actual: float
            The actual value that should be checked for not exceeding the Bound.
        """
        return actual > self.value

    def _check_upper_bound(self, actual: float) -> bool:
        """
        Check that a value is not higher or equal to the OpenBound.

        Parameters
        ----------
        actual: float
            The actual value that should be checked for not exceeding the Bound.
        """
        return actual < self.value



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from safeds.data.tabular.typing import Anything, Integer, Nothing, RealNumber
from safeds.data.tabular.typing._column_type import ColumnType
from safeds.exceptions import UnknownColumnNameError

if TYPE_CHECKING:
    import pandas as pd


@dataclass
class Schema:
    """
    Store column names and corresponding data types for a `Table` or `Row`.

    Parameters
    ----------
    schema : dict[str, ColumnType]
        Map from column names to data types.

    Examples
    --------
    >>> from safeds.data.tabular.typing import Integer, Schema, String
    >>> schema = Schema({"A": Integer(), "B": String()})
    """

    _schema: dict[str, ColumnType]

    # ------------------------------------------------------------------------------------------------------------------
    # Creation
    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def _from_pandas_dataframe(dataframe: pd.DataFrame) -> Schema:
        """
        Create a schema from a `pandas.DataFrame`.

        Parameters
        ----------
        dataframe : pd.DataFrame
            The dataframe.

        Returns
        -------
        schema : Schema
            The schema.
        """
        names = dataframe.columns
        # noinspection PyProtectedMember
        types = []
        for col in dataframe:
            types.append(ColumnType._data_type(dataframe[col]))

        return Schema(dict(zip(names, types, strict=True)))

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, schema: dict[str, ColumnType]):
        self._schema = dict(schema)  # Defensive copy

    def __hash__(self) -> int:
        """
        Return a hash value for the schema.

        Returns
        -------
        hash : int
            The hash value.

        Examples
        --------
        >>> from safeds.data.tabular.typing import Integer, Schema, String
        >>> schema = Schema({"A": Integer(), "B": String()})
        >>> hash_value = hash(schema)
        """
        column_names = self._schema.keys()
        column_types = map(repr, self._schema.values())
        return hash(tuple(zip(column_names, column_types, strict=True)))

    def __repr__(self) -> str:
        """
        Return an unambiguous string representation of this row.

        Returns
        -------
        representation : str
            The string representation.

        Examples
        --------
        >>> from safeds.data.tabular.typing import Integer, Schema, String
        >>> schema = Schema({"A": Integer()})
        >>> repr(schema)
        "Schema({'A': Integer})"
        """
        return f"Schema({self!s})"

    def __str__(self) -> str:
        """
        Return a user-friendly string representation of the schema.

        Returns
        -------
        string : str
            The string representation.

        Examples
        --------
        >>> from safeds.data.tabular.typing import Integer, Schema, String
        >>> schema = Schema({"A": Integer()})
        >>> str(schema)
        "{'A': Integer}"
        """
        match len(self._schema):
            case 0:
                return "{}"
            case 1:
                return str(self._schema)
            case _:
                lines = (f"    {name!r}: {type_}" for name, type_ in self._schema.items())
                joined = ",\n".join(lines)
                return f"{{\n{joined}\n}}"

    @property
    def column_names(self) -> list[str]:
        """
        Return a list of all column names saved in this schema.

        Returns
        -------
        column_names : list[str]
            The column names.

        Examples
        --------
        >>> from safeds.data.tabular.typing import Integer, Schema, String
        >>> schema = Schema({"A": Integer(), "B": String()})
        >>> schema.column_names
        ['A', 'B']
        """
        return list(self._schema.keys())

    def has_column(self, column_name: str) -> bool:
        """
        Return whether the schema contains a given column.

        Parameters
        ----------
        column_name : str
            The name of the column.

        Returns
        -------
        contains : bool
            True if the schema contains the column.

        Examples
        --------
        >>> from safeds.data.tabular.typing import Integer, Schema, String
        >>> schema = Schema({"A": Integer(), "B": String()})
        >>> schema.has_column("A")
        True

        >>> schema.has_column("C")
        False
        """
        return column_name in self._schema

    def get_column_type(self, column_name: str) -> ColumnType:
        """
        Return the type of the given column.

        Parameters
        ----------
        column_name : str
            The name of the column.

        Returns
        -------
        type : ColumnType
            The type of the column.

        Raises
        ------
        UnknownColumnNameError
            If the specified column name does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.typing import Integer, Schema, String
        >>> schema = Schema({"A": Integer(), "B": String()})
        >>> schema.get_column_type("A")
        Integer
        """
        if not self.has_column(column_name):
            raise UnknownColumnNameError([column_name])
        return self._schema[column_name]

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    def to_dict(self) -> dict[str, ColumnType]:
        """
        Return a dictionary that maps column names to column types.

        Returns
        -------
        data : dict[str, ColumnType]
            Dictionary representation of the schema.

        Examples
        --------
        >>> from safeds.data.tabular.typing import Integer, Schema, String
        >>> schema = Schema({"A": Integer(), "B": String()})
        >>> schema.to_dict()
        {'A': Integer, 'B': String}
        """
        return dict(self._schema)  # defensive copy

    @staticmethod
    def merge_multiple_schemas(schemas: list[Schema]) -> Schema:
        """
        Merge multiple schemas into one.

        For each type missmatch the new schema will have the least common supertype.

        The type hierarchy is as follows:
        * Anything
            * RealNumber
                * Integer
            * Boolean
            * String

        Parameters
        ----------
        schemas : list[Schema]
            the list of schemas you want to merge

        Returns
        -------
        schema : Schema
            the new merged schema

        Raises
        ------
        UnknownColumnNameError
            if not all schemas have the same column names
        """
        schema_dict = schemas[0]._schema
        missing_col_names = set()
        for schema in schemas:
            missing_col_names.update(set(schema.column_names) - set(schema_dict.keys()))
        if len(missing_col_names) > 0:
            raise UnknownColumnNameError(list(missing_col_names))
        for schema in schemas:
            if schema_dict != schema._schema:
                for col_name in schema_dict:
                    nullable = False
                    if schema_dict[col_name].is_nullable() or schema.get_column_type(col_name).is_nullable():
                        nullable = True
                    if isinstance(schema_dict[col_name], type(schema.get_column_type(col_name))):
                        if schema.get_column_type(col_name).is_nullable() and not schema_dict[col_name].is_nullable():
                            schema_dict[col_name] = type(schema.get_column_type(col_name))(nullable)
                        continue
                    if (
                        isinstance(schema_dict[col_name], RealNumber)
                        and isinstance(schema.get_column_type(col_name), Integer)
                    ) or (
                        isinstance(schema_dict[col_name], Integer)
                        and isinstance(schema.get_column_type(col_name), RealNumber)
                    ):
                        schema_dict[col_name] = RealNumber(nullable)
                        continue
                    if isinstance(schema_dict[col_name], Nothing):
                        schema_dict[col_name] = type(schema.get_column_type(col_name))(nullable)
                        continue
                    if isinstance(schema.get_column_type(col_name), Nothing):
                        schema_dict[col_name] = type(schema_dict[col_name])(nullable)
                        continue
                    schema_dict[col_name] = Anything(nullable)
        return Schema(schema_dict)

    # ------------------------------------------------------------------------------------------------------------------
    # IPython Integration
    # ------------------------------------------------------------------------------------------------------------------

    def _repr_markdown_(self) -> str:
        """
        Return a Markdown representation of the schema.

        Returns
        -------
        markdown : str
            The Markdown representation.
        """
        if len(self._schema) == 0:
            return "Empty Schema"

        lines = (f"| {name} | {type_} |" for name, type_ in self._schema.items())
        joined = "\n".join(lines)
        return f"| Column Name | Column Type |\n| --- | --- |\n{joined}"



from __future__ import annotations

import warnings
from typing import Any

import pandas as pd
from sklearn.impute import SimpleImputer as sk_SimpleImputer

from safeds.data.tabular.containers import Table
from safeds.data.tabular.transformation._table_transformer import TableTransformer
from safeds.data.tabular.typing import ImputerStrategy
from safeds.exceptions import NonNumericColumnError, TransformerNotFittedError, UnknownColumnNameError


class Imputer(TableTransformer):
    """
    Replace missing values with the given strategy.

    Parameters
    ----------
    strategy : ImputerStrategy
        The strategy used to impute missing values. Use the classes nested inside `Imputer.Strategy` to specify it.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Column, Table
    >>> from safeds.data.tabular.transformation import Imputer
    >>>
    >>> table = Table.from_columns(
    ...     [
    ...         Column("a", [1, 3, None]),
    ...         Column("b", [None, 2, 3]),
    ...     ],
    ... )
    >>> transformer = Imputer(Imputer.Strategy.Constant(0))
    >>> transformed_table = transformer.fit_and_transform(table)
    """

    class Strategy:
        class Constant(ImputerStrategy):
            """
            An imputation strategy for imputing missing data with given constant values.

            Parameters
            ----------
            value :
                The given value to impute missing values.
            """

            def __init__(self, value: Any):
                self._value = value

            def __str__(self) -> str:
                return f"Constant({self._value})"

            def _augment_imputer(self, imputer: sk_SimpleImputer) -> None:
                imputer.strategy = "constant"
                imputer.fill_value = self._value

        class Mean(ImputerStrategy):
            """An imputation strategy for imputing missing data with mean values."""

            def __str__(self) -> str:
                return "Mean"

            def _augment_imputer(self, imputer: sk_SimpleImputer) -> None:
                imputer.strategy = "mean"

        class Median(ImputerStrategy):
            """An imputation strategy for imputing missing data with median values."""

            def __str__(self) -> str:
                return "Median"

            def _augment_imputer(self, imputer: sk_SimpleImputer) -> None:
                imputer.strategy = "median"

        class Mode(ImputerStrategy):
            """An imputation strategy for imputing missing data with mode values. The lowest value will be used if there are multiple values with the same highest count."""

            def __str__(self) -> str:
                return "Mode"

            def _augment_imputer(self, imputer: sk_SimpleImputer) -> None:
                imputer.strategy = "most_frequent"

    def __init__(self, strategy: ImputerStrategy):
        self._strategy = strategy

        self._wrapped_transformer: sk_SimpleImputer | None = None
        self._column_names: list[str] | None = None

    # noinspection PyProtectedMember
    def fit(self, table: Table, column_names: list[str] | None) -> Imputer:
        """
        Learn a transformation for a set of columns in a table.

        This transformer is not modified.

        Parameters
        ----------
        table : Table
            The table used to fit the transformer.
        column_names : list[str] | None
            The list of columns from the table used to fit the transformer. If `None`, all columns are used.

        Returns
        -------
        fitted_transformer : TableTransformer
            The fitted transformer.

        Raises
        ------
        UnknownColumnNameError
            If column_names contain a column name that is missing in the table
        ValueError
            If the table contains 0 rows
        NonNumericColumnError
            If the strategy is set to either Mean or Median and the specified columns of the table contain non-numerical data.
        """
        if column_names is None:
            column_names = table.column_names
        else:
            missing_columns = sorted(set(column_names) - set(table.column_names))
            if len(missing_columns) > 0:
                raise UnknownColumnNameError(missing_columns)

        if table.number_of_rows == 0:
            raise ValueError("The Imputer cannot be fitted because the table contains 0 rows")

        if (isinstance(self._strategy, Imputer.Strategy.Mean | Imputer.Strategy.Median)) and table.keep_only_columns(
            column_names,
        ).remove_columns_with_non_numerical_values().number_of_columns < len(
            column_names,
        ):
            raise NonNumericColumnError(
                str(
                    sorted(
                        set(table.keep_only_columns(column_names).column_names)
                        - set(
                            table.keep_only_columns(column_names)
                            .remove_columns_with_non_numerical_values()
                            .column_names,
                        ),
                    ),
                ),
            )

        if isinstance(self._strategy, Imputer.Strategy.Mode):
            multiple_most_frequent = {}
            for name in column_names:
                if len(table.get_column(name).mode()) > 1:
                    multiple_most_frequent[name] = table.get_column(name).mode()
            if len(multiple_most_frequent) > 0:
                warnings.warn(
                    (
                        "There are multiple most frequent values in a column given to the Imputer.\nThe lowest values"
                        " are being chosen in this cases. The following columns have multiple most frequent"
                        f" values:\n{multiple_most_frequent}"
                    ),
                    UserWarning,
                    stacklevel=2,
                )

        wrapped_transformer = sk_SimpleImputer()
        self._strategy._augment_imputer(wrapped_transformer)
        wrapped_transformer.fit(table._data[column_names])

        result = Imputer(self._strategy)
        result._wrapped_transformer = wrapped_transformer
        result._column_names = column_names

        return result

    # noinspection PyProtectedMember
    def transform(self, table: Table) -> Table:
        """
        Apply the learned transformation to a table.

        The table is not modified.

        Parameters
        ----------
        table : Table
            The table to which the learned transformation is applied.

        Returns
        -------
        transformed_table : Table
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        UnknownColumnNameError
            If the input table does not contain all columns used to fit the transformer.
        ValueError
            If the table contains 0 rows.
        """
        # Transformer has not been fitted yet
        if self._wrapped_transformer is None or self._column_names is None:
            raise TransformerNotFittedError

        # Input table does not contain all columns used to fit the transformer
        missing_columns = sorted(set(self._column_names) - set(table.column_names))
        if len(missing_columns) > 0:
            raise UnknownColumnNameError(missing_columns)

        if table.number_of_rows == 0:
            raise ValueError("The Imputer cannot transform the table because it contains 0 rows")

        data = table._data.copy()
        data[self._column_names] = pd.DataFrame(
            self._wrapped_transformer.transform(data[self._column_names]),
            columns=self._column_names,
        )
        return Table._from_pandas_dataframe(data, table.schema)

    def is_fitted(self) -> bool:
        """
        Check if the transformer is fitted.

        Returns
        -------
        is_fitted : bool
            Whether the transformer is fitted.
        """
        return self._wrapped_transformer is not None

    def get_names_of_added_columns(self) -> list[str]:
        """
        Get the names of all new columns that have been added by the Imputer.

        Returns
        -------
        added_columns : list[str]
            A list of names of the added columns, ordered as they will appear in the table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        """
        if not self.is_fitted():
            raise TransformerNotFittedError
        return []

    # (Must implement abstract method, cannot instantiate class otherwise.)
    def get_names_of_changed_columns(self) -> list[str]:
        """
         Get the names of all columns that may have been changed by the Imputer.

        Returns
        -------
        changed_columns : list[str]
             The list of (potentially) changed column names, as passed to fit.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        """
        if self._column_names is None:
            raise TransformerNotFittedError
        return self._column_names

    def get_names_of_removed_columns(self) -> list[str]:
        """
        Get the names of all columns that have been removed by the Imputer.

        Returns
        -------
        removed_columns : list[str]
            A list of names of the removed columns, ordered as they appear in the table the Imputer was fitted on.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        """
        if not self.is_fitted():
            raise TransformerNotFittedError
        return []



from __future__ import annotations

import copy
import functools
import io
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

import Levenshtein
import matplotlib.pyplot as plt
import numpy as np
import openpyxl
import pandas as pd
import seaborn as sns
from pandas import DataFrame
from scipy import stats

from safeds.data.image.containers import Image
from safeds.data.image.typing import ImageFormat
from safeds.data.tabular.typing import ColumnType, Schema
from safeds.exceptions import (
    ColumnLengthMismatchError,
    ColumnSizeError,
    DuplicateColumnNameError,
    IndexOutOfBoundsError,
    NonNumericColumnError,
    UnknownColumnNameError,
    WrongFileExtensionError,
)

from ._column import Column
from ._row import Row

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence

    from safeds.data.tabular.transformation import InvertibleTableTransformer, TableTransformer

    from ._tagged_table import TaggedTable


# noinspection PyProtectedMember
class Table:
    """
    A table is a two-dimensional collection of data. It can either be seen as a list of rows or as a list of columns.

    To create a `Table` call the constructor or use one of the following static methods:

    | Method                                                                       | Description                            |
    | ---------------------------------------------------------------------------- | -------------------------------------- |
    | [from_csv_file][safeds.data.tabular.containers._table.Table.from_csv_file]   | Create a table from a CSV file.        |
    | [from_json_file][safeds.data.tabular.containers._table.Table.from_json_file] | Create a table from a JSON file.       |
    | [from_dict][safeds.data.tabular.containers._table.Table.from_dict]           | Create a table from a dictionary.      |
    | [from_columns][safeds.data.tabular.containers._table.Table.from_columns]     | Create a table from a list of columns. |
    | [from_rows][safeds.data.tabular.containers._table.Table.from_rows]           | Create a table from a list of rows.    |

    Note: When removing the last column of the table, the `number_of_columns` property will be set to 0.

    Parameters
    ----------
    data : Mapping[str, Sequence[Any]] | None
        The data. If None, an empty table is created.

    Raises
    ------
    ColumnLengthMismatchError
        If columns have different lengths.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Table
    >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Creation
    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def from_csv_file(path: str | Path) -> Table:
        """
        Read data from a CSV file into a table.

        Parameters
        ----------
        path : str | Path
            The path to the CSV file.

        Returns
        -------
        table : Table
            The table created from the CSV file.

        Raises
        ------
        FileNotFoundError
            If the specified file does not exist.
        WrongFileExtensionError
            If the file is not a csv file.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> Table.from_csv_file('./src/resources/from_csv_file.csv')
           a  b  c
        0  1  2  1
        1  0  0  7
        """
        path = Path(path)
        if path.suffix != ".csv":
            raise WrongFileExtensionError(path, ".csv")
        if path.exists():
            with path.open() as f:
                if f.read().replace("\n", "") == "":
                    return Table()

            return Table._from_pandas_dataframe(pd.read_csv(path))
        else:
            raise FileNotFoundError(f'File "{path}" does not exist')

    @staticmethod
    def from_excel_file(path: str | Path) -> Table:
        """
        Read data from an Excel file into a table.

        Valid file extensions are `.xls`, '.xlsx', `.xlsm`, `.xlsb`, `.odf`, `.ods` and `.odt`.

        Parameters
        ----------
        path : str | Path
            The path to the Excel file.

        Returns
        -------
        table : Table
            The table created from the Excel file.

        Raises
        ------
        FileNotFoundError
            If the specified file does not exist.
        WrongFileExtensionError
            If the file is not an Excel file.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> Table.from_excel_file('./src/resources/from_excel_file.xlsx')
           a  b
        0  1  4
        1  2  5
        2  3  6
        """
        path = Path(path)
        excel_extensions = [".xls", ".xlsx", ".xlsm", ".xlsb", ".odf", ".ods", ".odt"]
        if path.suffix not in excel_extensions:
            raise WrongFileExtensionError(path, excel_extensions)
        try:
            return Table._from_pandas_dataframe(
                pd.read_excel(path, engine="openpyxl", usecols=lambda colname: "Unnamed" not in colname),
            )
        except FileNotFoundError as exception:
            raise FileNotFoundError(f'File "{path}" does not exist') from exception

    @staticmethod
    def from_json_file(path: str | Path) -> Table:
        """
        Read data from a JSON file into a table.

        Parameters
        ----------
        path : str | Path
            The path to the JSON file.

        Returns
        -------
        table : Table
            The table created from the JSON file.

        Raises
        ------
        FileNotFoundError
            If the specified file does not exist.
        WrongFileExtensionError
            If the file is not a JSON file.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> Table.from_json_file('./src/resources/from_json_file.json')
           a  b
        0  1  4
        1  2  5
        2  3  6
        """
        path = Path(path)
        if path.suffix != ".json":
            raise WrongFileExtensionError(path, ".json")
        if path.exists():
            with path.open() as f:
                if f.read().replace("\n", "") in ("", "{}"):
                    return Table()

            return Table._from_pandas_dataframe(pd.read_json(path))
        else:
            raise FileNotFoundError(f'File "{path}" does not exist')

    @staticmethod
    def from_dict(data: dict[str, list[Any]]) -> Table:
        """
        Create a table from a dictionary that maps column names to column values.

        Parameters
        ----------
        data : dict[str, list[Any]]
            The data.

        Returns
        -------
        table : Table
            The generated table.

        Raises
        ------
        ColumnLengthMismatchError
            If columns have different lengths.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> d = {'a': [1, 2], 'b': [3, 4]}
        >>> Table.from_dict(d)
           a  b
        0  1  3
        1  2  4
        """
        return Table(data)

    @staticmethod
    def from_columns(columns: list[Column]) -> Table:
        """
        Return a table created from a list of columns.

        Parameters
        ----------
        columns : list[Column]
            The columns to be combined. They need to have the same size.

        Returns
        -------
        table : Table
            The generated table.

        Raises
        ------
        ColumnLengthMismatchError
            If any of the column sizes does not match with the others.
        DuplicateColumnNameError
            If multiple columns have the same name.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column, Table
        >>> col1 = Column("a", [1, 2, 3])
        >>> col2 = Column("b", [4, 5, 6])
        >>> Table.from_columns([col1, col2])
           a  b
        0  1  4
        1  2  5
        2  3  6
        """
        dataframe: DataFrame = pd.DataFrame()
        column_names = []

        for column in columns:
            if column._data.size != columns[0]._data.size:
                raise ColumnLengthMismatchError(
                    "\n".join(f"{column.name}: {column._data.size}" for column in columns),
                )
            if column.name in column_names:
                raise DuplicateColumnNameError(column.name)
            column_names.append(column.name)
            dataframe[column.name] = column._data

        return Table._from_pandas_dataframe(dataframe)

    @staticmethod
    def from_rows(rows: list[Row]) -> Table:
        """
        Return a table created from a list of rows.

        Parameters
        ----------
        rows : list[Row]
            The rows to be combined. They need to have a matching schema.

        Returns
        -------
        table : Table
            The generated table.

        Raises
        ------
        UnknownColumnNameError
            If any of the row column names does not match with the first row.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row, Table
        >>> row1 = Row({"a": 1, "b": 2})
        >>> row2 = Row({"a": 3, "b": 4})
        >>> Table.from_rows([row1, row2])
           a  b
        0  1  2
        1  3  4
        """
        if len(rows) == 0:
            return Table._from_pandas_dataframe(pd.DataFrame())

        column_names_compare: list = list(rows[0].column_names)
        unknown_column_names = set()
        row_array: list[pd.DataFrame] = []

        for row in rows:
            unknown_column_names.update(set(column_names_compare) - set(row.column_names))
            row_array.append(row._data)
        if len(unknown_column_names) > 0:
            raise UnknownColumnNameError(list(unknown_column_names))

        dataframe: DataFrame = pd.concat(row_array, ignore_index=True)
        dataframe.columns = column_names_compare

        schema = Schema.merge_multiple_schemas([row.schema for row in rows])

        return Table._from_pandas_dataframe(dataframe, schema)

    @staticmethod
    def _from_pandas_dataframe(data: pd.DataFrame, schema: Schema | None = None) -> Table:
        """
        Create a table from a `pandas.DataFrame`.

        Parameters
        ----------
        data : pd.DataFrame
            The data.
        schema : Schema | None
            The schema. If None, the schema is inferred from the data.

        Returns
        -------
        table : Table
            The created table.

        Examples
        --------
        >>> import pandas as pd
        >>> from safeds.data.tabular.containers import Table
        >>> Table._from_pandas_dataframe(pd.DataFrame({"a": [1], "b": [2]}))
           a  b
        0  1  2
        """
        data = data.reset_index(drop=True)

        result = object.__new__(Table)
        result._data = data

        if schema is None:
            # noinspection PyProtectedMember
            result._schema = Schema._from_pandas_dataframe(data)
        else:
            result._schema = schema
            if result._data.empty:
                result._data = pd.DataFrame(columns=schema.column_names)

        return result

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, data: Mapping[str, Sequence[Any]] | None = None) -> None:
        """
        Create a table from a mapping of column names to their values.

        Parameters
        ----------
        data : Mapping[str, Sequence[Any]] | None
            The data. If None, an empty table is created.

        Raises
        ------
        ColumnLengthMismatchError
            If columns have different lengths.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> Table({"a": [1, 2, 3], "b": [4, 5, 6]})
           a  b
        0  1  4
        1  2  5
        2  3  6
        """
        if data is None:
            data = {}

        # Validation
        expected_length: int | None = None
        for column_values in data.values():
            if expected_length is None:
                expected_length = len(column_values)
            elif len(column_values) != expected_length:
                raise ColumnLengthMismatchError(
                    "\n".join(f"{column_name}: {len(column_values)}" for column_name, column_values in data.items()),
                )

        # Implementation
        self._data: pd.DataFrame = pd.DataFrame(data)
        self._data = self._data.reset_index(drop=True)
        self._schema: Schema = Schema._from_pandas_dataframe(self._data)

    def __eq__(self, other: Any) -> bool:
        """
        Compare two table instances.

        Returns
        -------
        'True' if contents are equal, 'False' otherwise.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row, Table
        >>> row1 = Row({"a": 1, "b": 2})
        >>> row2 = Row({"a": 3, "b": 4})
        >>> row3 = Row({"a": 5, "b": 6})
        >>> table1 = Table.from_rows([row1, row2])
        >>> table2 = Table.from_rows([row1, row2])
        >>> table3 = Table.from_rows([row1, row3])
        >>> table1 == table2
        True
        >>> table1 == table3
        False
        """
        if not isinstance(other, Table):
            return NotImplemented
        if self is other:
            return True
        if self.number_of_columns == 0 and other.number_of_columns == 0:
            return True
        table1 = self.sort_columns()
        table2 = other.sort_columns()
        if table1.number_of_rows == 0 and table2.number_of_rows == 0:
            return table1.column_names == table2.column_names
        return table1._schema == table2._schema and table1._data.equals(table2._data)

    def __repr__(self) -> str:
        r"""
        Display the table in only one line.

        Returns
        -------
        A string representation of the table in only one line.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3], "b": [2, 4]})
        >>> repr(table)
        '   a  b\n0  1  2\n1  3  4'
        """
        tmp = self._data.copy(deep=True)
        tmp.columns = self.column_names
        return tmp.__repr__()

    def __str__(self) -> str:
        tmp = self._data.copy(deep=True)
        tmp.columns = self.column_names
        return tmp.__str__()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def column_names(self) -> list[str]:
        """
        Return a list of all column names in this table.

        Alias for self.schema.column_names -> list[str].

        Returns
        -------
        column_names : list[str]
            The list of the column names.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"col1": [1, 3], "col2": [2, 4]})
        >>> table.column_names
        ['col1', 'col2']
        """
        return self._schema.column_names

    @property
    def number_of_columns(self) -> int:
        """
        Return the number of columns.

        Returns
        -------
        number_of_columns : int
            The number of columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1], "b": [2]})
        >>> table.number_of_columns
        2
        """
        return self._data.shape[1]

    @property
    def number_of_rows(self) -> int:
        """
        Return the number of rows.

        Returns
        -------
        number_of_rows : int
            The number of rows.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1], "b": [2]})
        >>> table.number_of_rows
        1
        """
        return self._data.shape[0]

    @property
    def schema(self) -> Schema:
        """
        Return the schema of the table.

        Returns
        -------
        schema : Schema
            The schema.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row, Table
        >>> row = Row({"a": 1, "b": 2.5, "c": "ff"})
        >>> table = Table.from_dict({"a": [1, 8], "b": [2.5, 9], "c": ["g", "g"]})
        >>> table.schema
        Schema({
            'a': Integer,
            'b': RealNumber,
            'c': String
        })
        >>> table.schema == row.schema
        True
        """
        return self._schema

    # ------------------------------------------------------------------------------------------------------------------
    # Getters
    # ------------------------------------------------------------------------------------------------------------------

    def get_column(self, column_name: str) -> Column:
        """
        Return a column with the data of the specified column.

        Parameters
        ----------
        column_name : str
            The name of the column.

        Returns
        -------
        column : Column
            The column.

        Raises
        ------
        UnknownColumnNameError
            If the specified target column name does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1], "b": [2]})
        >>> table.get_column("b")
        Column('b', [2])
        """
        if not self.has_column(column_name):
            similar_columns = self._get_similar_columns(column_name)
            raise UnknownColumnNameError([column_name], similar_columns)

        return Column._from_pandas_series(
            self._data[column_name],
            self.get_column_type(column_name),
        )

    def has_column(self, column_name: str) -> bool:
        """
        Return whether the table contains a given column.

        Alias for self.schema.hasColumn(column_name: str) -> bool.

        Parameters
        ----------
        column_name : str
            The name of the column.

        Returns
        -------
        contains : bool
            True if the column exists.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1], "b": [2]})
        >>> table.has_column("b")
        True
        >>> table.has_column("c")
        False
        """
        return self._schema.has_column(column_name)

    def get_column_type(self, column_name: str) -> ColumnType:
        """
        Return the type of the given column.

        Alias for self.schema.get_type_of_column(column_name: str) -> ColumnType.

        Parameters
        ----------
        column_name : str
            The name of the column to be queried.

        Returns
        -------
        type : ColumnType
            The type of the column.

        Raises
        ------
        UnknownColumnNameError
            If the specified target column name does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1], "b": [2.5]})
        >>> table.get_column_type("b")
        RealNumber
        """
        return self._schema.get_column_type(column_name)

    def get_row(self, index: int) -> Row:
        """
        Return the row at a specified index.

        Parameters
        ----------
        index : int
            The index.

        Returns
        -------
        row : Row
            The row of the table at the index.

        Raises
        ------
        IndexOutOfBoundsError
            If no row at the specified index exists in this table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3], "b": [2, 4]})
        >>> table.get_row(0)
        Row({
            'a': 1,
            'b': 2
        })
        """
        if len(self._data.index) - 1 < index or index < 0:
            raise IndexOutOfBoundsError(index)

        return Row._from_pandas_dataframe(self._data.iloc[[index]], self._schema)

    def _get_similar_columns(self, column_name: str) -> list[str]:
        """
        Get all the column names in a Table that are similar to a given name.

        Parameters
        ----------
        column_name : str
            The name to compare the Table's column names to.

        Returns
        -------
        similar_columns: list[str]
            A list of all column names in the Table that are similar or equal to the given column name.
        """
        similar_columns = []
        similarity = 0.6
        i = 0
        while i < len(self.column_names):
            if Levenshtein.jaro_winkler(self.column_names[i], column_name) >= similarity:
                similar_columns.append(self.column_names[i])
            i += 1
            if len(similar_columns) == 4 and similarity < 0.9:
                similarity += 0.1
                similar_columns = []
                i = 0

        return similar_columns

    # ------------------------------------------------------------------------------------------------------------------
    # Information
    # ------------------------------------------------------------------------------------------------------------------

    def summarize_statistics(self) -> Table:
        """
        Return a table with a number of statistical key values.

        The original table is not modified.

        Returns
        -------
        result : Table
            The table with statistics.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3], "b": [2, 4]})
        >>> table.summarize_statistics()
                      metrics                   a                   b
        0             maximum                   3                   4
        1             minimum                   1                   2
        2                mean                 2.0                 3.0
        3                mode              [1, 3]              [2, 4]
        4              median                 2.0                 3.0
        5                 sum                   4                   6
        6            variance                 2.0                 2.0
        7  standard deviation  1.4142135623730951  1.4142135623730951
        8              idness                 1.0                 1.0
        9           stability                 0.5                 0.5
        """
        if self.number_of_columns == 0:
            return Table(
                {
                    "metrics": [
                        "maximum",
                        "minimum",
                        "mean",
                        "mode",
                        "median",
                        "sum",
                        "variance",
                        "standard deviation",
                        "idness",
                        "stability",
                    ],
                },
            )
        elif self.number_of_rows == 0:
            table = Table(
                {
                    "metrics": [
                        "maximum",
                        "minimum",
                        "mean",
                        "mode",
                        "median",
                        "sum",
                        "variance",
                        "standard deviation",
                        "idness",
                        "stability",
                    ],
                },
            )
            for name in self.column_names:
                table = table.add_column(Column(name, ["-", "-", "-", "-", "-", "-", "-", "-", "-", "-"]))
            return table

        columns = self.to_columns()
        result = pd.DataFrame()
        statistics = {}

        for column in columns:
            statistics = {
                "maximum": column.maximum,
                "minimum": column.minimum,
                "mean": column.mean,
                "mode": column.mode,
                "median": column.median,
                "sum": column.sum,
                "variance": column.variance,
                "standard deviation": column.standard_deviation,
                "idness": column.idness,
                "stability": column.stability,
            }
            values = []

            for function in statistics.values():
                try:
                    values.append(str(function()))
                except (NonNumericColumnError, ValueError):
                    values.append("-")

            result = pd.concat([result, pd.DataFrame(values)], axis=1)

        result = pd.concat([pd.DataFrame(list(statistics.keys())), result], axis=1)
        result.columns = ["metrics", *self.column_names]

        return Table._from_pandas_dataframe(result)

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    # This method is meant as a way to "cast" instances of subclasses of `Table` to a proper `Table`, dropping any
    # additional constraints that might have to hold in the subclass.
    # Override accordingly in subclasses.
    def _as_table(self: Table) -> Table:
        """
        Transform the table to an instance of the Table class.

        Returns
        -------
        table: Table
            The table, as an instance of the Table class.
        """
        return self

    def add_column(self, column: Column) -> Table:
        """
        Return a new table with the provided column attached at the end.

        The original table is not modified.

        Returns
        -------
        result : Table
            The table with the column attached.

        Raises
        ------
        DuplicateColumnNameError
            If the new column already exists.
        ColumnSizeError
            If the size of the column does not match the number of rows.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3], "b": [2, 4]})
        >>> col = Column("c", ["d", "e"])
        >>> table.add_column(col)
           a  b  c
        0  1  2  d
        1  3  4  e
        """
        if self.has_column(column.name):
            raise DuplicateColumnNameError(column.name)

        if column.number_of_rows != self.number_of_rows and self.number_of_columns != 0:
            raise ColumnSizeError(str(self.number_of_rows), str(column._data.size))

        result = self._data.copy()
        result.columns = self._schema.column_names
        result[column.name] = column._data
        return Table._from_pandas_dataframe(result)

    def add_columns(self, columns: list[Column] | Table) -> Table:
        """
        Return a new `Table` with multiple added columns.

        The original table is not modified.

        Parameters
        ----------
        columns : list[Column] or Table
            The columns to be added.

        Returns
        -------
        result: Table
            A new table combining the original table and the given columns.

        Raises
        ------
        DuplicateColumnNameError
            If at least one column name from the provided column list already exists in the table.
        ColumnSizeError
            If at least one of the column sizes from the provided column list does not match the table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column, Table
        >>> table = Table.from_dict({"a": [1, 3], "b": [2, 4]})
        >>> col1 = Column("c", ["d", "e"])
        >>> col2 = Column("d", [3.5, 7.9])
        >>> table.add_columns([col1, col2])
           a  b  c    d
        0  1  2  d  3.5
        1  3  4  e  7.9
        """
        if isinstance(columns, Table):
            columns = columns.to_columns()
        result = self._data.copy()
        result.columns = self._schema.column_names
        for column in columns:
            if column.name in result.columns:
                raise DuplicateColumnNameError(column.name)

            if column.number_of_rows != self.number_of_rows and self.number_of_columns != 0:
                raise ColumnSizeError(str(self.number_of_rows), str(column._data.size))

            result[column.name] = column._data
        return Table._from_pandas_dataframe(result)

    def add_row(self, row: Row) -> Table:
        """
        Return a new `Table` with an added Row attached.

        If the table happens to be empty beforehand, respective columns will be added automatically.

        The order of columns of the new row will be adjusted to the order of columns in the table.
        The new table will contain the merged schema.

        The original table is not modified.

        Parameters
        ----------
        row : Row
            The row to be added.

        Returns
        -------
        table : Table
            A new table with the added row at the end.

        Raises
        ------
        UnknownColumnNameError
            If the row has different column names than the table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row, Table
        >>> table = Table.from_dict({"a": [1], "b": [2]})
        >>> row = Row.from_dict({"a": 3, "b": 4})
        >>> table.add_row(row)
           a  b
        0  1  2
        1  3  4
        """
        int_columns = []
        result = self._copy()
        if self.number_of_columns == 0:
            return Table.from_rows([row])
        if len(set(self.column_names) - set(row.column_names)) > 0:
            raise UnknownColumnNameError(
                sorted(
                    set(self.column_names) - set(row.column_names),
                    key={val: ix for ix, val in enumerate(self.column_names)}.__getitem__,
                ),
            )

        if result.number_of_rows == 0:
            int_columns = list(filter(lambda name: isinstance(row[name], int | np.int64 | np.int32), row.column_names))

        new_df = pd.concat([result._data, row._data]).infer_objects()
        new_df.columns = result.column_names
        schema = Schema.merge_multiple_schemas([result.schema, row.schema])
        result = Table._from_pandas_dataframe(new_df, schema)

        for column in int_columns:
            result = result.replace_column(column, [result.get_column(column).transform(lambda it: int(it))])

        return result

    def add_rows(self, rows: list[Row] | Table) -> Table:
        """
        Return a new `Table` with multiple added Rows attached.

        The order of columns of the new rows will be adjusted to the order of columns in the table.
        The new table will contain the merged schema.

        The original table is not modified.

        Parameters
        ----------
        rows : list[Row] or Table
            The rows to be added.

        Returns
        -------
        result : Table
            A new table which combines the original table and the given rows.

        Raises
        ------
        UnknownColumnNameError
            If at least one of the rows have different column names than the table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Row, Table
        >>> table = Table.from_dict({"a": [1], "b": [2]})
        >>> row1 = Row.from_dict({"a": 3, "b": 4})
        >>> row2 = Row.from_dict({"a": 5, "b": 6})
        >>> table.add_rows([row1, row2])
           a  b
        0  1  2
        1  3  4
        2  5  6
        """
        if isinstance(rows, Table):
            rows = rows.to_rows()

        if len(rows) == 0:
            return self._copy()

        different_column_names = set()
        for row in rows:
            different_column_names.update(set(self.column_names) - set(row.column_names))
        if len(different_column_names) > 0:
            raise UnknownColumnNameError(
                sorted(
                    different_column_names,
                    key={val: ix for ix, val in enumerate(self.column_names)}.__getitem__,
                ),
            )

        result = self._copy()

        for row in rows:
            result = result.add_row(row)

        return result

    def filter_rows(self, query: Callable[[Row], bool]) -> Table:
        """
        Return a new table with rows filtered by Callable (e.g. lambda function).

        The original table is not modified.

        Parameters
        ----------
        query : lambda function
            A Callable that is applied to all rows.

        Returns
        -------
        table : Table
            A table containing only the rows filtered by the query.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3], "b": [2, 4]})
        >>> table.filter_rows(lambda x: x["a"] < 2)
           a  b
        0  1  2
        """
        rows: list[Row] = [row for row in self.to_rows() if query(row)]
        if len(rows) == 0:
            result_table = Table._from_pandas_dataframe(pd.DataFrame(), self._schema)
        else:
            result_table = self.from_rows(rows)
        return result_table

    _T = TypeVar("_T")

    def group_rows_by(self, key_selector: Callable[[Row], _T]) -> dict[_T, Table]:
        """
        Return a dictionary with copies of the output tables as values and the keys from the key_selector.

        The original table is not modified.

        Parameters
        ----------
        key_selector : Callable[[Row], _T]
            A Callable that is applied to all rows and returns the key of the group.

        Returns
        -------
        dictionary : dict
            A dictionary containing the new tables as values and the selected keys as keys.
        """
        dictionary: dict[Table._T, Table] = {}
        for row in self.to_rows():
            if key_selector(row) in dictionary:
                dictionary[key_selector(row)] = dictionary[key_selector(row)].add_row(row)
            else:
                dictionary[key_selector(row)] = Table.from_rows([row])
        return dictionary

    def keep_only_columns(self, column_names: list[str]) -> Table:
        """
        Return a new table with only the given column(s).

        The original table is not modified.

        Note: When removing the last column of the table, the `number_of_columns` property will be set to 0.

        Parameters
        ----------
        column_names : list[str]
            A list containing only the columns to be kept.

        Returns
        -------
        table : Table
            A table containing only the given column(s).

        Raises
        ------
        UnknownColumnNameError
            If any of the given columns does not exist.
        IllegalSchemaModificationError
            If removing the columns would violate an invariant in the subclass.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3], "b": [2, 4]})
        >>> table.keep_only_columns(["b"])
           b
        0  2
        1  4
        """
        invalid_columns = []
        similar_columns: list[str] = []
        for name in column_names:
            if not self._schema.has_column(name):
                similar_columns = similar_columns + self._get_similar_columns(name)
                invalid_columns.append(name)
        if len(invalid_columns) != 0:
            raise UnknownColumnNameError(invalid_columns, similar_columns)

        clone = self._copy()
        clone = clone.remove_columns(list(set(self.column_names) - set(column_names)))
        return clone

    def remove_columns(self, column_names: list[str]) -> Table:
        """
        Return a new table without the given column(s).

        The original table is not modified.

        Note: When removing the last column of the table, the `number_of_columns` property will be set to 0.

        Parameters
        ----------
        column_names : list[str]
            A list containing all columns to be dropped.

        Returns
        -------
        table : Table
            A table without the given columns.

        Raises
        ------
        UnknownColumnNameError
            If any of the given columns does not exist.
        IllegalSchemaModificationError
            If removing the columns would violate an invariant in the subclass.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3], "b": [2, 4]})
        >>> table.remove_columns(["b"])
           a
        0  1
        1  3
        """
        invalid_columns = []
        similar_columns: list[str] = []
        for name in column_names:
            if not self._schema.has_column(name):
                similar_columns = similar_columns + self._get_similar_columns(name)
                invalid_columns.append(name)
        if len(invalid_columns) != 0:
            raise UnknownColumnNameError(invalid_columns, similar_columns)

        transformed_data = self._data.drop(labels=column_names, axis="columns")
        transformed_data.columns = [name for name in self._schema.column_names if name not in column_names]

        if len(transformed_data.columns) == 0:
            return Table()

        return Table._from_pandas_dataframe(transformed_data)

    def remove_columns_with_missing_values(self) -> Table:
        """
        Return a new table without the columns that contain missing values.

        The original table is not modified.

        Note: When removing the last column of the table, the `number_of_columns` property will be set to 0.

        Returns
        -------
        table : Table
            A table without the columns that contain missing values.

        Raises
        ------
        IllegalSchemaModificationError
            If removing the columns would violate an invariant in the subclass.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 2], "b": [None, 2]})
        >>> table.remove_columns_with_missing_values()
           a
        0  1
        1  2
        """
        return Table.from_columns([column for column in self.to_columns() if not column.has_missing_values()])

    def remove_columns_with_non_numerical_values(self) -> Table:
        """
        Return a new table without the columns that contain non-numerical values.

        The original table is not modified.

        Note: When removing the last column of the table, the `number_of_columns` property will be set to 0.

        Returns
        -------
        table : Table
            A table without the columns that contain non-numerical values.

        Raises
        ------
        IllegalSchemaModificationError
            If removing the columns would violate an invariant in the subclass.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 0], "b": ["test", 2]})
        >>> table.remove_columns_with_non_numerical_values()
           a
        0  1
        1  0
        """
        return Table.from_columns([column for column in self.to_columns() if column.type.is_numeric()])

    def remove_duplicate_rows(self) -> Table:
        """
        Return a new table with every duplicate row removed.

        The original table is not modified.

        Returns
        -------
        result : Table
            The table with the duplicate rows removed.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3, 3], "b": [2, 4, 4]})
        >>> table.remove_duplicate_rows()
           a  b
        0  1  2
        1  3  4
        """
        result = self._data.drop_duplicates(ignore_index=True)
        result.columns = self._schema.column_names
        return Table._from_pandas_dataframe(result)

    def remove_rows_with_missing_values(self) -> Table:
        """
        Return a new table without the rows that contain missing values.

        The original table is not modified.

        Returns
        -------
        table : Table
            A table without the rows that contain missing values.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1.0, None, 3], "b": [2, 4.0, None]})
        >>> table.remove_rows_with_missing_values()
             a    b
        0  1.0  2.0
        """
        result = self._data.copy(deep=True)
        result = result.dropna(axis="index")
        return Table._from_pandas_dataframe(result)

    def remove_rows_with_outliers(self) -> Table:
        """
        Return a new table without those rows that contain at least one outlier.

        We define an outlier as a value that has a distance of more than 3 standard deviations from the column mean.
        Missing values are not considered outliers. They are also ignored during the calculation of the standard
        deviation.

        The original table is not modified.

        Returns
        -------
        new_table : Table
            A new table without rows containing outliers.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column, Table
        >>> c1 = Column("a", [1, 3, 1, 0.1, 0, 0, 0, 0, 0, 0, 0, 0])
        >>> c2 = Column("b", [1.5, 1, 0.5, 0.01, 0, 0, 0, 0, 0, 0, 0, 0])
        >>> c3 = Column("c", [0.1, 0.00, 0.4, 0.2, 0, 0, 0, 0, 0, 0, 0, 0])
        >>> c4 = Column("d", [-1000000, 1000000, -1000000, -1000000, -1000000, -1000000, -1000000, -1000000, -1000000, -1000000, -1000000, -1000000])
        >>> table = Table.from_columns([c1, c2, c3, c4])
        >>> table.remove_rows_with_outliers()
              a     b    c        d
        0   1.0  1.50  0.1 -1000000
        1   1.0  0.50  0.4 -1000000
        2   0.1  0.01  0.2 -1000000
        3   0.0  0.00  0.0 -1000000
        4   0.0  0.00  0.0 -1000000
        5   0.0  0.00  0.0 -1000000
        6   0.0  0.00  0.0 -1000000
        7   0.0  0.00  0.0 -1000000
        8   0.0  0.00  0.0 -1000000
        9   0.0  0.00  0.0 -1000000
        10  0.0  0.00  0.0 -1000000
        """
        copy = self._data.copy(deep=True)

        table_without_nonnumericals = self.remove_columns_with_non_numerical_values()
        z_scores = np.absolute(stats.zscore(table_without_nonnumericals._data, nan_policy="omit"))
        filter_ = ((z_scores < 3) | np.isnan(z_scores)).all(axis=1)

        return Table._from_pandas_dataframe(copy[filter_], self._schema)

    def rename_column(self, old_name: str, new_name: str) -> Table:
        """
        Return a new `Table` with a single column renamed.

        The original table is not modified.

        Parameters
        ----------
        old_name : str
            The old name of the target column.
        new_name : str
            The new name of the target column.

        Returns
        -------
        table : Table
            The Table with the renamed column.

        Raises
        ------
        UnknownColumnNameError
            If the specified old target column name does not exist.
        DuplicateColumnNameError
            If the specified new target column name already exists.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1], "b": [2]})
        >>> table.rename_column("b", "c")
           a  c
        0  1  2
        """
        if old_name not in self._schema.column_names:
            similar_columns = self._get_similar_columns(old_name)
            raise UnknownColumnNameError([old_name], similar_columns)
        if old_name == new_name:
            return self
        if new_name in self._schema.column_names:
            raise DuplicateColumnNameError(new_name)

        new_df = self._data.copy()
        new_df.columns = self._schema.column_names
        return Table._from_pandas_dataframe(new_df.rename(columns={old_name: new_name}))

    def replace_column(self, old_column_name: str, new_columns: list[Column]) -> Table:
        """
        Return a new table with the specified old column replaced by a list of new columns.

        The order of columns is kept.

        The original table is not modified.

        Parameters
        ----------
        old_column_name : str
            The name of the column to be replaced.

        new_columns : list[Column]
            The list of new columns replacing the old column.

        Returns
        -------
        result : Table
            A table with the old column replaced by the new columns.

        Raises
        ------
        UnknownColumnNameError
            If the old column does not exist.
        DuplicateColumnNameError
            If at least one of the new columns already exists and the existing column is not affected by the replacement.
        ColumnSizeError
            If the size of at least one of the new columns does not match the amount of rows.
        IllegalSchemaModificationError
            If replacing the column would violate an invariant in the subclass.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column, Table
        >>> table = Table.from_dict({"a": [1], "b": [2]})
        >>> new_col = Column("new", [3])
        >>> table.replace_column("b", [new_col])
           a  new
        0  1    3
        """
        if old_column_name not in self._schema.column_names:
            similar_columns = self._get_similar_columns(old_column_name)
            raise UnknownColumnNameError([old_column_name], similar_columns)

        columns = list[Column]()
        for old_column in self.column_names:
            if old_column == old_column_name:
                for new_column in new_columns:
                    if new_column.name in self.column_names and new_column.name != old_column_name:
                        raise DuplicateColumnNameError(new_column.name)

                    if self.number_of_rows != new_column.number_of_rows:
                        raise ColumnSizeError(str(self.number_of_rows), str(new_column.number_of_rows))
                    columns.append(new_column)
            else:
                columns.append(self.get_column(old_column))

        return Table.from_columns(columns)

    def shuffle_rows(self) -> Table:
        """
        Return a new `Table` with randomly shuffled rows of this `Table`.

        The original table is not modified.

        Returns
        -------
        result : Table
            The shuffled Table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> import numpy as np
        >>> np.random.seed(123456)
        >>> table = Table.from_dict({"a": [1, 3, 5], "b": [2, 4, 6]})
        >>> table.shuffle_rows()
           a  b
        0  5  6
        1  1  2
        2  3  4
        """
        new_df = self._data.sample(frac=1.0)
        new_df.columns = self._schema.column_names
        return Table._from_pandas_dataframe(new_df)

    def slice_rows(
        self,
        start: int | None = None,
        end: int | None = None,
        step: int = 1,
    ) -> Table:
        """
        Slice a part of the table into a new table.

        The original table is not modified.

        Parameters
        ----------
        start : int | None
            The first index of the range to be copied into a new table, None by default.
        end : int | None
            The last index of the range to be copied into a new table, None by default.
        step : int
            The step size used to iterate through the table, 1 by default.

        Returns
        -------
        result : Table
            The resulting table.

        Raises
        ------
        IndexOutOfBoundsError
            If the index is out of bounds.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3, 5], "b": [2, 4, 6]})
        >>> table.slice_rows(0, 2)
           a  b
        0  1  2
        1  3  4
        """
        if start is None:
            start = 0

        if end is None:
            end = self.number_of_rows

        if end < start:
            raise IndexOutOfBoundsError(slice(start, end))
        if start < 0 or end < 0 or start > self.number_of_rows or end > self.number_of_rows:
            raise IndexOutOfBoundsError(start if start < 0 or start > self.number_of_rows else end)

        new_df = self._data.iloc[start:end:step]
        new_df.columns = self._schema.column_names
        return Table._from_pandas_dataframe(new_df)

    def sort_columns(
        self,
        comparator: Callable[[Column, Column], int] = lambda col1, col2: (col1.name > col2.name)
        - (col1.name < col2.name),
    ) -> Table:
        """
        Sort the columns of a `Table` with the given comparator and return a new `Table`.

        The comparator is a function that takes two columns `col1` and `col2` and
        returns an integer:

        * If `col1` should be ordered before `col2`, the function should return a negative number.
        * If `col1` should be ordered after `col2`, the function should return a positive number.
        * If the original order of `col1` and `col2` should be kept, the function should return 0.

        If no comparator is given, the columns will be sorted alphabetically by their name.

        The original table is not modified.

        Parameters
        ----------
        comparator : Callable[[Column, Column], int]
            The function used to compare two columns.

        Returns
        -------
        new_table : Table
            A new table with sorted columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1], "b": [2] })
        >>> table.sort_columns(lambda col1, col2: 1)
           a  b
        0  1  2
        >>> table.sort_columns(lambda col1, col2: -1)
           b  a
        0  2  1
        >>> table2 = Table.from_dict({"b": [2], "a": [1]})
        >>> table2.sort_columns()
           a  b
        0  1  2
        """
        columns = self.to_columns()
        columns.sort(key=functools.cmp_to_key(comparator))
        return Table.from_columns(columns)

    def sort_rows(self, comparator: Callable[[Row, Row], int]) -> Table:
        """
        Sort the rows of a `Table` with the given comparator and return a new `Table`.

        The comparator is a function that takes two rows `row1` and `row2` and
        returns an integer:

        * If `row1` should be ordered before `row2`, the function should return a negative number.
        * If `row1` should be ordered after `row2`, the function should return a positive number.
        * If the original order of `row1` and `row2` should be kept, the function should return 0.

        The original table is not modified.

        Parameters
        ----------
        comparator : Callable[[Row, Row], int]
            The function used to compare two rows.

        Returns
        -------
        new_table : Table
            A new table with sorted rows.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 3, 5], "b": [2, 4, 6] })
        >>> table.sort_rows(lambda row1, row2: 1)
           a  b
        0  1  2
        1  3  4
        2  5  6
        >>> table.sort_rows(lambda row1, row2: -1)
           a  b
        0  5  6
        1  3  4
        2  1  2
        >>> table.sort_rows(lambda row1, row2: 0)
           a  b
        0  1  2
        1  3  4
        2  5  6
        """
        rows = self.to_rows()
        rows.sort(key=functools.cmp_to_key(comparator))
        return Table.from_rows(rows)

    def split_rows(self, percentage_in_first: float) -> tuple[Table, Table]:
        """
        Split the table into two new tables.

        The original table is not modified.

        Parameters
        ----------
        percentage_in_first : float
            The desired size of the first table in percentage to the given table; must be between 0 and 1.

        Returns
        -------
        result : (Table, Table)
            A tuple containing the two resulting tables. The first table has the specified size, the second table
            contains the rest of the data.

        Raises
        ------
        ValueError:
            if the 'percentage_in_first' is not between 0 and 1.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"temperature": [10, 15, 20, 25, 30], "sales": [54, 74, 90, 206, 210]})
        >>> slices = table.split_rows(0.4)
        >>> slices[0]
           temperature  sales
        0           10     54
        1           15     74
        >>> slices[1]
           temperature  sales
        0           20     90
        1           25    206
        2           30    210
        """
        if percentage_in_first < 0 or percentage_in_first > 1:
            raise ValueError("The given percentage is not between 0 and 1")
        if self.number_of_rows == 0:
            return Table(), Table()
        return (
            self.slice_rows(0, round(percentage_in_first * self.number_of_rows)),
            self.slice_rows(round(percentage_in_first * self.number_of_rows)),
        )

    def tag_columns(self, target_name: str, feature_names: list[str] | None = None) -> TaggedTable:
        """
        Return a new `TaggedTable` with columns marked as a target column or feature columns.

        The original table is not modified.

        Parameters
        ----------
        target_name : str
            Name of the target column.
        feature_names : list[str] | None
            Names of the feature columns. If None, all columns except the target column are used.

        Returns
        -------
        tagged_table : TaggedTable
            A new tagged table with the given target and feature names.

        Raises
        ------
        ValueError
            If the target column is also a feature column.
        ValueError
            If no feature columns are specified.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table, TaggedTable
        >>> table = Table.from_dict({"item": ["apple", "milk", "beer"], "price": [1.10, 1.19, 1.79], "amount_bought": [74, 72, 51]})
        >>> tagged_table = table.tag_columns(target_name="amount_bought", feature_names=["item", "price"])
        """
        from ._tagged_table import TaggedTable

        return TaggedTable._from_table(self, target_name, feature_names)

    def transform_column(self, name: str, transformer: Callable[[Row], Any]) -> Table:
        """
        Return a new `Table` with the provided column transformed by calling the provided transformer.

        The original table is not modified.

        Returns
        -------
        result : Table
            The table with the transformed column.

        Raises
        ------
        UnknownColumnNameError
            If the column does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"item": ["apple", "milk", "beer"], "price": [1.00, 1.19, 1.79]})
        >>> table.transform_column("price", lambda row: row.get_value("price") * 100)
            item  price
        0  apple  100.0
        1   milk  119.0
        2   beer  179.0
        """
        if self.has_column(name):
            items: list = [transformer(item) for item in self.to_rows()]
            result: list[Column] = [Column(name, items)]
            return self.replace_column(name, result)
        similar_columns = self._get_similar_columns(name)
        raise UnknownColumnNameError([name], similar_columns)

    def transform_table(self, transformer: TableTransformer) -> Table:
        """
        Return a new `Table` with a learned transformation applied to this table.

        The original table is not modified.

        Parameters
        ----------
        transformer : TableTransformer
            The transformer which transforms the given table.

        Returns
        -------
        transformed_table : Table
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        IllegalSchemaModificationError
            If replacing the column would violate an invariant in the subclass.

        Examples
        --------
        >>> from safeds.data.tabular.transformation import OneHotEncoder
        >>> from safeds.data.tabular.containers import Table
        >>> transformer = OneHotEncoder()
        >>> table = Table.from_dict({"fruit": ["apple", "pear", "apple"], "pet": ["dog", "duck", "duck"]})
        >>> transformer = transformer.fit(table, None)
        >>> table.transform_table(transformer)
           fruit__apple  fruit__pear  pet__dog  pet__duck
        0           1.0          0.0       1.0        0.0
        1           0.0          1.0       0.0        1.0
        2           1.0          0.0       0.0        1.0
        """
        return transformer.transform(self)

    def inverse_transform_table(self, transformer: InvertibleTableTransformer) -> Table:
        """
        Return a new `Table` with the inverted transformation applied by the given transformer.

        The original table is not modified.

        Parameters
        ----------
        transformer : InvertibleTableTransformer
            A transformer that was fitted with columns, which are all present in the table.

        Returns
        -------
        table : Table
            The original table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.

        Examples
        --------
        >>> from safeds.data.tabular.transformation import OneHotEncoder
        >>> from safeds.data.tabular.containers import Table
        >>> transformer = OneHotEncoder()
        >>> table = Table.from_dict({"a": ["j", "k", "k"], "b": ["x", "y", "x"]})
        >>> transformer = transformer.fit(table, None)
        >>> transformed_table = transformer.transform(table)
        >>> transformed_table.inverse_transform_table(transformer)
           a  b
        0  j  x
        1  k  y
        2  k  x
        >>> transformer.inverse_transform(transformed_table)
           a  b
        0  j  x
        1  k  y
        2  k  x
        """
        return transformer.inverse_transform(self)

    # ------------------------------------------------------------------------------------------------------------------
    # Plotting
    # ------------------------------------------------------------------------------------------------------------------

    def plot_correlation_heatmap(self) -> Image:
        """
        Plot a correlation heatmap for all numerical columns of this `Table`.

        Returns
        -------
        plot: Image
            The plot as an image.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"temperature": [10, 15, 20, 25, 30], "sales": [54, 74, 90, 206, 210]})
        >>> image = table.plot_correlation_heatmap()
        """
        only_numerical = self.remove_columns_with_non_numerical_values()

        if self.number_of_rows == 0:
            warnings.warn(
                "An empty table has been used. A correlation heatmap on an empty table will show nothing.",
                stacklevel=2,
            )

            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    message=(
                        "Attempting to set identical low and high (xlims|ylims) makes transformation singular;"
                        " automatically expanding."
                    ),
                )
                fig = plt.figure()
                sns.heatmap(
                    data=only_numerical._data.corr(),
                    vmin=-1,
                    vmax=1,
                    xticklabels=only_numerical.column_names,
                    yticklabels=only_numerical.column_names,
                    cmap="vlag",
                )
                plt.tight_layout()
        else:
            fig = plt.figure()
            sns.heatmap(
                data=only_numerical._data.corr(),
                vmin=-1,
                vmax=1,
                xticklabels=only_numerical.column_names,
                yticklabels=only_numerical.column_names,
                cmap="vlag",
            )
            plt.tight_layout()

        buffer = io.BytesIO()
        fig.savefig(buffer, format="png")
        plt.close()  # Prevents the figure from being displayed directly
        buffer.seek(0)
        return Image(buffer, format_=ImageFormat.PNG)

    def plot_lineplot(self, x_column_name: str, y_column_name: str) -> Image:
        """
        Plot two columns against each other in a lineplot.

        If there are multiple x-values for a y-value, the resulting plot will consist of a line representing the mean
        and the lower-transparency area around the line representing the 95% confidence interval.

        Parameters
        ----------
        x_column_name : str
            The column name of the column to be plotted on the x-Axis.
        y_column_name : str
            The column name of the column to be plotted on the y-Axis.

        Returns
        -------
        plot: Image
            The plot as an image.

        Raises
        ------
        UnknownColumnNameError
            If either of the columns do not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"temperature": [10, 15, 20, 25, 30], "sales": [54, 74, 90, 206, 210]})
        >>> image = table.plot_lineplot("temperature", "sales")
        """
        if not self.has_column(x_column_name) or not self.has_column(y_column_name):
            similar_columns_x = self._get_similar_columns(x_column_name)
            similar_columns_y = self._get_similar_columns(y_column_name)
            raise UnknownColumnNameError(
                ([x_column_name] if not self.has_column(x_column_name) else [])
                + ([y_column_name] if not self.has_column(y_column_name) else []),
                (similar_columns_x if not self.has_column(x_column_name) else [])
                + (similar_columns_y if not self.has_column(y_column_name) else []),
            )

        fig = plt.figure()
        ax = sns.lineplot(
            data=self._data,
            x=x_column_name,
            y=y_column_name,
        )
        ax.set(xlabel=x_column_name, ylabel=y_column_name)
        ax.set_xticks(ax.get_xticks())
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            horizontalalignment="right",
        )  # rotate the labels of the x Axis to prevent the chance of overlapping of the labels
        plt.tight_layout()

        buffer = io.BytesIO()
        fig.savefig(buffer, format="png")
        plt.close()  # Prevents the figure from being displayed directly
        buffer.seek(0)
        return Image(buffer, format_=ImageFormat.PNG)

    def plot_scatterplot(self, x_column_name: str, y_column_name: str) -> Image:
        """
        Plot two columns against each other in a scatterplot.

        Parameters
        ----------
        x_column_name : str
            The column name of the column to be plotted on the x-Axis.
        y_column_name : str
            The column name of the column to be plotted on the y-Axis.

        Returns
        -------
        plot: Image
            The plot as an image.

        Raises
        ------
        UnknownColumnNameError
            If either of the columns do not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"temperature": [10, 15, 20, 25, 30], "sales": [54, 74, 90, 206, 210]})
        >>> image = table.plot_scatterplot("temperature", "sales")
        """
        if not self.has_column(x_column_name) or not self.has_column(y_column_name):
            similar_columns_x = self._get_similar_columns(x_column_name)
            similar_columns_y = self._get_similar_columns(y_column_name)
            raise UnknownColumnNameError(
                ([x_column_name] if not self.has_column(x_column_name) else [])
                + ([y_column_name] if not self.has_column(y_column_name) else []),
                (similar_columns_x if not self.has_column(x_column_name) else [])
                + (similar_columns_y if not self.has_column(y_column_name) else []),
            )

        fig = plt.figure()
        ax = sns.scatterplot(
            data=self._data,
            x=x_column_name,
            y=y_column_name,
        )
        ax.set(xlabel=x_column_name, ylabel=y_column_name)
        ax.set_xticks(ax.get_xticks())
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            horizontalalignment="right",
        )  # rotate the labels of the x Axis to prevent the chance of overlapping of the labels
        plt.tight_layout()

        buffer = io.BytesIO()
        fig.savefig(buffer, format="png")
        plt.close()  # Prevents the figure from being displayed directly
        buffer.seek(0)
        return Image(buffer, format_=ImageFormat.PNG)

    def plot_boxplots(self) -> Image:
        """
        Plot a boxplot for every numerical column.

        Returns
        -------
        plot: Image
            The plot as an image.

        Raises
        ------
        NonNumericColumnError
            If the table contains only non-numerical columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a":[1, 2], "b": [3, 42]})
        >>> image = table.plot_boxplots()
        """
        numerical_table = self.remove_columns_with_non_numerical_values()
        if numerical_table.number_of_columns == 0:
            raise NonNumericColumnError("This table contains only non-numerical columns.")
        col_wrap = min(numerical_table.number_of_columns, 3)

        data = pd.melt(numerical_table._data, value_vars=numerical_table.column_names)
        grid = sns.FacetGrid(data, col="variable", col_wrap=col_wrap, sharex=False, sharey=False)
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="Using the boxplot function without specifying `order` is likely to produce an incorrect plot.",
            )
            grid.map(sns.boxplot, "variable", "value")
        grid.set_xlabels("")
        grid.set_ylabels("")
        grid.set_titles("{col_name}")
        for axes in grid.axes.flat:
            axes.set_xticks([])
        plt.tight_layout()
        fig = grid.fig

        buffer = io.BytesIO()
        fig.savefig(buffer, format="png")
        plt.close()  # Prevents the figure from being displayed directly
        buffer.seek(0)
        return Image(buffer, format_=ImageFormat.PNG)

    def plot_histograms(self) -> Image:
        """
        Plot a histogram for every column.

        Returns
        -------
        plot: Image
            The plot as an image.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [2, 3, 5, 1], "b": [54, 74, 90, 2014]})
        >>> image = table.plot_histograms()
        """
        col_wrap = min(self.number_of_columns, 3)

        data = pd.melt(self._data.applymap(lambda value: str(value)), value_vars=self.column_names)
        grid = sns.FacetGrid(data=data, col="variable", col_wrap=col_wrap, sharex=False, sharey=False)
        grid.map(sns.histplot, "value")
        grid.set_xlabels("")
        grid.set_ylabels("")
        grid.set_titles("{col_name}")
        for axes in grid.axes.flat:
            axes.set_xticks(axes.get_xticks())
            axes.set_xticklabels(axes.get_xticklabels(), rotation=45, horizontalalignment="right")
        grid.tight_layout()
        fig = grid.fig

        buffer = io.BytesIO()
        fig.savefig(buffer, format="png")
        plt.close()
        buffer.seek(0)
        return Image(buffer, ImageFormat.PNG)

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    def to_csv_file(self, path: str | Path) -> None:
        """
        Write the data from the table into a CSV file.

        If the file and/or the directories do not exist they will be created. If the file already exists it will be
        overwritten.

        Parameters
        ----------
        path : str | Path
            The path to the output file.

        Raises
        ------
        WrongFileExtensionError
            If the file is not a csv file.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.to_csv_file("./src/resources/to_csv_file.csv")
        """
        path = Path(path)
        if path.suffix != ".csv":
            raise WrongFileExtensionError(path, ".csv")
        path.parent.mkdir(parents=True, exist_ok=True)
        data_to_csv = self._data.copy()
        data_to_csv.columns = self._schema.column_names
        data_to_csv.to_csv(path, index=False)

    def to_excel_file(self, path: str | Path) -> None:
        """
        Write the data from the table into an Excel file.

        Valid file extensions are `.xls`, '.xlsx', `.xlsm`, `.xlsb`, `.odf`, `.ods` and `.odt`.
        If the file and/or the directories do not exist, they will be created. If the file already exists, it will be
        overwritten.

        Parameters
        ----------
        path : str | Path
            The path to the output file.

        Raises
        ------
        WrongFileExtensionError
            If the file is not an Excel file.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.to_excel_file("./src/resources/to_excel_file.xlsx")
        """
        path = Path(path)
        excel_extensions = [".xls", ".xlsx", ".xlsm", ".xlsb", ".odf", ".ods", ".odt"]
        if path.suffix not in excel_extensions:
            raise WrongFileExtensionError(path, excel_extensions)

        # Create Excel metadata in the file
        tmp_table_file = openpyxl.Workbook()
        tmp_table_file.save(path)

        path.parent.mkdir(parents=True, exist_ok=True)
        data_to_excel = self._data.copy()
        data_to_excel.columns = self._schema.column_names
        data_to_excel.to_excel(path)

    def to_json_file(self, path: str | Path) -> None:
        """
        Write the data from the table into a JSON file.

        If the file and/or the directories do not exist, they will be created. If the file already exists it will be
        overwritten.

        Parameters
        ----------
        path : str | Path
            The path to the output file.

        Raises
        ------
        WrongFileExtensionError
            If the file is not a JSON file.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.to_json_file("./src/resources/to_json_file.json")
        """
        path = Path(path)
        if path.suffix != ".json":
            raise WrongFileExtensionError(path, ".json")
        path.parent.mkdir(parents=True, exist_ok=True)
        data_to_json = self._data.copy()
        data_to_json.columns = self._schema.column_names
        data_to_json.to_json(path)

    def to_dict(self) -> dict[str, list[Any]]:
        """
        Return a dictionary that maps column names to column values.

        Returns
        -------
        data : dict[str, list[Any]]
            Dictionary representation of the table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> row1 = Row({"a": 1, "b": 5})
        >>> row2 = Row({"a": 2, "b": 6})
        >>> table1 = Table.from_rows([row1, row2])
        >>> table2 = Table.from_dict({"a": [1, 2], "b": [5, 6]})
        >>> table1 == table2
        True
        """
        return {column_name: list(self.get_column(column_name)) for column_name in self.column_names}

    def to_html(self) -> str:
        """
        Return an HTML representation of the table.

        Returns
        -------
        output : str
            The generated HTML.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> html = table.to_html()
        """
        return self._data.to_html(max_rows=self._data.shape[0], max_cols=self._data.shape[1])

    def to_columns(self) -> list[Column]:
        """
        Return a list of the columns.

        Returns
        -------
        columns : list[Columns]
            List of columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a":[1, 2],"b":[20, 30]})
        >>> table.to_columns()
        [Column('a', [1, 2]), Column('b', [20, 30])]
        """
        return [self.get_column(name) for name in self._schema.column_names]

    def to_rows(self) -> list[Row]:
        """
        Return a list of the rows.

        Returns
        -------
        rows : list[Row]
            List of rows.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table.from_dict({"a":[1, 2],"b":[20, 30]})
        >>> table.to_rows()
        [Row({
            'a': 1,
            'b': 20
        }), Row({
            'a': 2,
            'b': 30
        })]
        """
        return [
            Row._from_pandas_dataframe(
                pd.DataFrame([list(series_row)], columns=self._schema.column_names),
                self._schema,
            )
            for (_, series_row) in self._data.iterrows()
        ]

    # ------------------------------------------------------------------------------------------------------------------
    # IPython integration
    # ------------------------------------------------------------------------------------------------------------------

    def _repr_html_(self) -> str:
        """
        Return an HTML representation of the table.

        Returns
        -------
        output : str
            The generated HTML.
        """
        return self._data.to_html(max_rows=self._data.shape[0], max_cols=self._data.shape[1], notebook=True)

    # ------------------------------------------------------------------------------------------------------------------
    # Dataframe interchange protocol
    # ------------------------------------------------------------------------------------------------------------------

    def __dataframe__(self, nan_as_null: bool = False, allow_copy: bool = True):  # type: ignore[no-untyped-def]
        """
        Return a DataFrame exchange object that conforms to the dataframe interchange protocol.

        Generally, there is no reason to call this method directly. The dataframe interchange protocol is designed to
        allow libraries to consume tabular data from different sources, such as `pandas` or `polars`. If you still
        decide to call this method, you should not rely on any capabilities of the returned object beyond the dataframe
        interchange protocol.

        The specification of the dataframe interchange protocol can be found on
        [GitHub](https://github.com/data-apis/dataframe-api).

        Parameters
        ----------
        nan_as_null : bool
            Whether to replace missing values in the data with `NaN`.
        allow_copy : bool
            Whether memory may be copied to create the DataFrame exchange object.

        Returns
        -------
        dataframe
            A DataFrame object that conforms to the dataframe interchange protocol.
        """
        if not allow_copy:
            raise NotImplementedError("For the moment we need to copy the data, so `allow_copy` must be True.")

        data_copy = self._data.copy()
        data_copy.columns = self.column_names
        return data_copy.__dataframe__(nan_as_null, allow_copy)

    # ------------------------------------------------------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------------------------------------------------------

    def _copy(self) -> Table:
        """
        Return a copy of this table.

        Returns
        -------
        table : Table
            The copy of this table.
        """
        return copy.deepcopy(self)



"""Models for regression tasks."""

from ._ada_boost import AdaBoost
from ._decision_tree import DecisionTree
from ._elastic_net_regression import ElasticNetRegression
from ._gradient_boosting import GradientBoosting
from ._k_nearest_neighbors import KNearestNeighbors
from ._lasso_regression import LassoRegression
from ._linear_regression import LinearRegression
from ._random_forest import RandomForest
from ._regressor import Regressor
from ._ridge_regression import RidgeRegression
from ._support_vector_machine import SupportVectorMachine

__all__ = [
    "AdaBoost",
    "DecisionTree",
    "ElasticNetRegression",
    "GradientBoosting",
    "KNearestNeighbors",
    "LassoRegression",
    "LinearRegression",
    "RandomForest",
    "Regressor",
    "RidgeRegression",
    "SupportVectorMachine",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from sklearn.tree import DecisionTreeClassifier as sk_DecisionTreeClassifier

from safeds.ml.classical._util_sklearn import fit, predict

from ._classifier import Classifier

if TYPE_CHECKING:
    from sklearn.base import ClassifierMixin

    from safeds.data.tabular.containers import Table, TaggedTable


class DecisionTree(Classifier):
    """Decision tree classification."""

    def __init__(self) -> None:
        # Internal state
        self._wrapped_classifier: sk_DecisionTreeClassifier | None = None
        self._feature_names: list[str] | None = None
        self._target_name: str | None = None

    def fit(self, training_set: TaggedTable) -> DecisionTree:
        """
        Create a copy of this classifier and fit it with the given training data.

        This classifier is not modified.

        Parameters
        ----------
        training_set : TaggedTable
            The training data containing the feature and target vectors.

        Returns
        -------
        fitted_classifier : DecisionTree
            The fitted classifier.

        Raises
        ------
        LearningError
            If the training data contains invalid values or if the training failed.
        UntaggedTableError
            If the table is untagged.
        NonNumericColumnError
            If the training data contains non-numerical values.
        MissingValuesColumnError
            If the training data contains missing values.
        DatasetMissesDataError
            If the training data contains no rows.
        """
        wrapped_classifier = self._get_sklearn_classifier()
        fit(wrapped_classifier, training_set)

        result = DecisionTree()
        result._wrapped_classifier = wrapped_classifier
        result._feature_names = training_set.features.column_names
        result._target_name = training_set.target.name

        return result

    def predict(self, dataset: Table) -> TaggedTable:
        """
        Predict a target vector using a dataset containing feature vectors. The model has to be trained first.

        Parameters
        ----------
        dataset : Table
            The dataset containing the feature vectors.

        Returns
        -------
        table : TaggedTable
            A dataset containing the given feature vectors and the predicted target vector.

        Raises
        ------
        ModelNotFittedError
            If the model has not been fitted yet.
        DatasetContainsTargetError
            If the dataset contains the target column already.
        DatasetMissesFeaturesError
            If the dataset misses feature columns.
        PredictionError
            If predicting with the given dataset failed.
        NonNumericColumnError
            If the dataset contains non-numerical values.
        MissingValuesColumnError
            If the dataset contains missing values.
        DatasetMissesDataError
            If the dataset contains no rows.
        """
        return predict(self._wrapped_classifier, dataset, self._feature_names, self._target_name)

    def is_fitted(self) -> bool:
        """
        Check if the classifier is fitted.

        Returns
        -------
        is_fitted : bool
            Whether the classifier is fitted.
        """
        return self._wrapped_classifier is not None

    def _get_sklearn_classifier(self) -> ClassifierMixin:
        return sk_DecisionTreeClassifier()



from __future__ import annotations

import copy
import io
from collections.abc import Sequence
from numbers import Number
from typing import TYPE_CHECKING, Any, TypeVar, overload

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from safeds.data.image.containers import Image
from safeds.data.image.typing import ImageFormat
from safeds.data.tabular.typing import ColumnType
from safeds.exceptions import (
    ColumnLengthMismatchError,
    ColumnSizeError,
    IndexOutOfBoundsError,
    NonNumericColumnError,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

T = TypeVar("T")
R = TypeVar("R")


class Column(Sequence[T]):
    """
    A column is a named collection of values.

    Parameters
    ----------
    name : str
        The name of the column.
    data : Sequence[T]
        The data.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Column
    >>> column = Column("test", [1, 2, 3])
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Creation
    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def _from_pandas_series(data: pd.Series, type_: ColumnType | None = None) -> Column:
        """
        Create a column from a `pandas.Series`.

        Parameters
        ----------
        data : pd.Series
            The data.
        type_ : ColumnType | None
            The type. If None, the type is inferred from the data.

        Returns
        -------
        column : Column
            The created column.

        Examples
        --------
        >>> import pandas as pd
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column._from_pandas_series(pd.Series([1, 2, 3], name="test"))
        """
        result = object.__new__(Column)
        result._name = data.name
        result._data = data
        # noinspection PyProtectedMember
        result._type = type_ if type_ is not None else ColumnType._data_type(data)

        return result

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, name: str, data: Sequence[T] | None = None) -> None:
        """
        Create a column.

        Parameters
        ----------
        name : str
            The name of the column.
        data : Sequence[T] | None
            The data. If None, an empty column is created.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        """
        if data is None:
            data = []

        self._name: str = name
        self._data: pd.Series = data.rename(name) if isinstance(data, pd.Series) else pd.Series(data, name=name)
        # noinspection PyProtectedMember
        self._type: ColumnType = ColumnType._data_type(self._data)

    def __contains__(self, item: Any) -> bool:
        return item in self._data

    def __eq__(self, other: object) -> bool:
        """
        Check whether this column is equal to another object.

        Parameters
        ----------
        other : object
            The other object.

        Returns
        -------
        equal : bool
            True if the other object is an identical column. False if the other object is a different column.
            NotImplemented if the other object is not a column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3])
        >>> column2 = Column("test", [1, 2, 3])
        >>> column1 == column2
        True

        >>> column3 = Column("test", [3, 4, 5])
        >>> column1 == column3
        False
        """
        if not isinstance(other, Column):
            return NotImplemented
        if self is other:
            return True
        return self.name == other.name and self._data.equals(other._data)

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, index: slice) -> Column[T]:
        ...

    def __getitem__(self, index: int | slice) -> T | Column[T]:
        """
        Return the value of the specified row or rows.

        Parameters
        ----------
        index : int | slice
            The index of the row, or a slice specifying the start and end index.

        Returns
        -------
        value : Any
            The single row's value, or rows' values.

        Raises
        ------
        IndexOutOfBoundsError
            If the given index or indices do not exist in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column[0]
        1
        """
        if isinstance(index, int):
            if index < 0 or index >= self._data.size:
                raise IndexOutOfBoundsError(index)
            return self._data[index]

        if isinstance(index, slice):
            if index.start < 0 or index.start > self._data.size:
                raise IndexOutOfBoundsError(index)
            if index.stop < 0 or index.stop > self._data.size:
                raise IndexOutOfBoundsError(index)
            data = self._data[index].reset_index(drop=True).rename(self.name)
            return Column._from_pandas_series(data, self._type)

    def __iter__(self) -> Iterator[T]:
        r"""
        Create an iterator for the data of this column. This way e.g. for-each loops can be used on it.

        Returns
        -------
        iterator : Iterator[Any]
            The iterator.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", ["A", "B", "C"])
        >>> string = ""
        >>> for val in column:
        ...     string += val + ", "
        >>> string
        'A, B, C, '
        """
        return iter(self._data)

    def __len__(self) -> int:
        """
        Return the size of the column.

        Returns
        -------
        n_rows : int
            The size of the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> len(column)
        3
        """
        return len(self._data)

    def __repr__(self) -> str:
        """
        Return an unambiguous string representation of this column.

        Returns
        -------
        representation : str
            The string representation.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> repr(column)
        "Column('test', [1, 2, 3])"
        """
        return f"Column({self._name!r}, {list(self._data)!r})"

    def __str__(self) -> str:
        """
        Return a user-friendly string representation of this column.

        Returns
        -------
        representation : str
            The string representation.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> str(column)
        "'test': [1, 2, 3]"
        """
        return f"{self._name!r}: {list(self._data)!r}"

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def name(self) -> str:
        """
        Return the name of the column.

        Returns
        -------
        name : str
            The name of the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.name
        'test'
        """
        return self._name

    @property
    def number_of_rows(self) -> int:
        """
        Return the number of elements in the column.

        Returns
        -------
        number_of_rows : int
            The number of elements.
        """
        return len(self._data)

    @property
    def type(self) -> ColumnType:
        """
        Return the type of the column.

        Returns
        -------
        type : ColumnType
            The type of the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.type
        Integer

        >>> column = Column("test", ['a', 'b', 'c'])
        >>> column.type
        String
        """
        return self._type

    # ------------------------------------------------------------------------------------------------------------------
    # Getters
    # ------------------------------------------------------------------------------------------------------------------

    def get_unique_values(self) -> list[T]:
        """
        Return a list of all unique values in the column.

        Returns
        -------
        unique_values : list[T]
            List of unique values in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3, 2, 4, 3])
        >>> column.get_unique_values()
        [1, 2, 3, 4]
        """
        return list(self._data.unique())

    def get_value(self, index: int) -> T:
        """
        Return column value at specified index, starting at 0.

        Parameters
        ----------
        index : int
            Index of requested element.

        Returns
        -------
        value
            Value at index in column.

        Raises
        ------
        IndexOutOfBoundsError
            If the given index does not exist in the column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.get_value(1)
        2
        """
        if index < 0 or index >= self._data.size:
            raise IndexOutOfBoundsError(index)

        return self._data[index]

    # ------------------------------------------------------------------------------------------------------------------
    # Information
    # ------------------------------------------------------------------------------------------------------------------

    def all(self, predicate: Callable[[T], bool]) -> bool:
        """
        Check if all values have a given property.

        Parameters
        ----------
        predicate : Callable[[T], bool])
            Callable that is used to find matches.

        Returns
        -------
        result : bool
            True if all match.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.all(lambda x: x < 4)
        True

        >>> column.all(lambda x: x < 2)
        False
        """
        return all(predicate(value) for value in self._data)

    def any(self, predicate: Callable[[T], bool]) -> bool:
        """
        Check if any value has a given property.

        Parameters
        ----------
        predicate : Callable[[T], bool])
            Callable that is used to find matches.

        Returns
        -------
        result : bool
            True if any match.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.any(lambda x: x < 2)
        True

        >>> column.any(lambda x: x < 1)
        False
        """
        return any(predicate(value) for value in self._data)

    def none(self, predicate: Callable[[T], bool]) -> bool:
        """
        Check if no values has a given property.

        Parameters
        ----------
        predicate : Callable[[T], bool])
            Callable that is used to find matches.

        Returns
        -------
        result : bool
            True if none match.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3])
        >>> column1.none(lambda x: x < 1)
        True

        >>> column2 = Column("test", [1, 2, 3])
        >>> column2.none(lambda x: x > 1)
        False
        """
        return all(not predicate(value) for value in self._data)

    def has_missing_values(self) -> bool:
        """
        Return whether the column has missing values.

        Returns
        -------
        missing_values_exist : bool
            True if missing values exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3, None])
        >>> column1.has_missing_values()
        True

        >>> column2 = Column("test", [1, 2, 3])
        >>> column2.has_missing_values()
        False
        """
        return self.any(lambda value: value is None or (isinstance(value, Number) and np.isnan(value)))

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    def rename(self, new_name: str) -> Column:
        """
        Return a new column with a new name.

        The original column is not modified.

        Parameters
        ----------
        new_name : str
            The new name of the column.

        Returns
        -------
        column : Column
            A new column with the new name.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.rename("new name")
        Column('new name', [1, 2, 3])
        """
        return Column._from_pandas_series(self._data.rename(new_name), self._type)

    def transform(self, transformer: Callable[[T], R]) -> Column[R]:
        """
        Apply a transform method to every data point.

        The original column is not modified.

        Parameters
        ----------
        transformer : Callable[[T], R]
            Function that will be applied to all data points.

        Returns
        -------
        transformed_column: Column
            The transformed column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> price = Column("price", [4.99, 5.99, 2.49])
        >>> sale = price.transform(lambda amount: amount * 0.8)
        """
        return Column(self.name, self._data.apply(transformer, convert_dtype=True))

    # ------------------------------------------------------------------------------------------------------------------
    # Statistics
    # ------------------------------------------------------------------------------------------------------------------

    def correlation_with(self, other_column: Column) -> float:
        """
        Calculate Pearson correlation between this and another column. Both columns have to be numerical.

        Returns
        -------
        correlation : float
            Correlation between the two columns.

        Raises
        ------
        NonNumericColumnError
            If one of the columns is not numerical.
        ColumnLengthMismatchError
            If the columns have different lengths.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3])
        >>> column2 = Column("test", [2, 4, 6])
        >>> column1.correlation_with(column2)
        1.0

        >>> column1 = Column("test", [1, 2, 3])
        >>> column2 = Column("test", [0.5, 4, -6])
        >>> column1.correlation_with(column2)
        -0.6404640308067906
        """
        if not self._type.is_numeric() or not other_column._type.is_numeric():
            raise NonNumericColumnError(
                f"Columns must be numerical. {self.name} is {self._type}, {other_column.name} is {other_column._type}.",
            )
        if self._data.size != other_column._data.size:
            raise ColumnLengthMismatchError(
                f"{self.name} is of size {self._data.size}, {other_column.name} is of size {other_column._data.size}.",
            )
        return self._data.corr(other_column._data)

    def idness(self) -> float:
        r"""
        Calculate the idness of this column.

        We define the idness as follows:

        $$
        \frac{\text{number of different values}}{\text{number of rows}}
        $$

        Returns
        -------
        idness : float
            The idness of the column.

        Raises
        ------
        ColumnSizeError
            If this column is empty.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3])
        >>> column1.idness()
        1.0

        >>> column2 = Column("test", [1, 2, 3, 2])
        >>> column2.idness()
        0.75
        """
        if self._data.size == 0:
            raise ColumnSizeError("> 0", "0")
        return self._data.nunique() / self._data.size

    def maximum(self) -> float:
        """
        Return the maximum value of the column. The column has to be numerical.

        Returns
        -------
        max : float
            The maximum value.

        Raises
        ------
        NonNumericColumnError
            If the data contains non-numerical data.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.maximum()
        3
        """
        if not self._type.is_numeric():
            raise NonNumericColumnError(f"{self.name} is of type {self._type}.")
        return self._data.max()

    def mean(self) -> float:
        """
        Return the mean value of the column. The column has to be numerical.

        Returns
        -------
        mean : float
            The mean value.

        Raises
        ------
        NonNumericColumnError
            If the data contains non-numerical data.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.mean()
        2.0
        """
        if not self._type.is_numeric():
            raise NonNumericColumnError(f"{self.name} is of type {self._type}.")
        return self._data.mean()

    def median(self) -> float:
        """
        Return the median value of the column. The column has to be numerical.

        Returns
        -------
        median : float
            The median value.

        Raises
        ------
        NonNumericColumnError
            If the data contains non-numerical data.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3, 4])
        >>> column.median()
        2.5

        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3, 4, 5])
        >>> column.median()
        3.0
        """
        if not self._type.is_numeric():
            raise NonNumericColumnError(f"{self.name} is of type {self._type}.")
        return self._data.median()

    def minimum(self) -> float:
        """
        Return the minimum value of the column. The column has to be numerical.

        Returns
        -------
        min : float
            The minimum value.

        Raises
        ------
        NonNumericColumnError
            If the data contains non-numerical data.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3, 4])
        >>> column.minimum()
        1
        """
        if not self._type.is_numeric():
            raise NonNumericColumnError(f"{self.name} is of type {self._type}.")
        return self._data.min()

    def missing_value_ratio(self) -> float:
        """
        Return the ratio of missing values to the total number of elements in the column.

        Returns
        -------
        ratio : float
            The ratio of missing values to the total number of elements in the column.

        Raises
        ------
        ColumnSizeError
            If the column is empty.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3, 4])
        >>> column1.missing_value_ratio()
        0.0

        >>> column2 = Column("test", [1, 2, 3, None])
        >>> column2.missing_value_ratio()
        0.25
        """
        if self._data.size == 0:
            raise ColumnSizeError("> 0", "0")
        return self._count_missing_values() / self._data.size

    def mode(self) -> list[T]:
        """
        Return the mode of the column.

        Returns
        -------
        mode: list[T]
            Returns a list with the most common values.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3, 3, 4])
        >>> column1.mode()
        [3]

        >>> column2 = Column("test", [1, 2, 3, 3, 4, 4])
        >>> column2.mode()
        [3, 4]
        """
        return self._data.mode().tolist()

    def stability(self) -> float:
        r"""
        Calculate the stability of this column.

        We define the stability as follows:

        $$
        \frac{\text{number of occurrences of most common non-null value}}{\text{number of non-null values}}
        $$

        The stability is not definded for a column with only null values.

        Returns
        -------
        stability : float
            The stability of the column.

        Raises
        ------
        ColumnSizeError
            If the column is empty.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 1, 2, 3])
        >>> column1.stability()
        0.5

        >>> column2 = Column("test", [1, 2, 2, 2, 3])
        >>> column2.stability()
        0.6
        """
        if self._data.size == 0:
            raise ColumnSizeError("> 0", "0")

        if self.all(lambda x: x is None):
            raise ValueError("Stability is not definded for a column with only null values.")

        return self._data.value_counts()[self.mode()[0]] / self._data.count()

    def standard_deviation(self) -> float:
        """
        Return the standard deviation of the column. The column has to be numerical.

        Returns
        -------
        sum : float
            The standard deviation of all values.

        Raises
        ------
        NonNumericColumnError
            If the data contains non-numerical data.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column1 = Column("test", [1, 2, 3])
        >>> column1.standard_deviation()
        1.0

        >>> column2 = Column("test", [1, 2, 4, 8, 16])
        >>> column2.standard_deviation()
        6.099180272790763
        """
        if not self.type.is_numeric():
            raise NonNumericColumnError(f"{self.name} is of type {self._type}.")
        return self._data.std()

    def sum(self) -> float:
        """
        Return the sum of the column. The column has to be numerical.

        Returns
        -------
        sum : float
            The sum of all values.

        Raises
        ------
        NonNumericColumnError
            If the data contains non-numerical data.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.sum()
        6
        """
        if not self.type.is_numeric():
            raise NonNumericColumnError(f"{self.name} is of type {self._type}.")
        return self._data.sum()

    def variance(self) -> float:
        """
        Return the variance of the column. The column has to be numerical.

        Returns
        -------
        sum : float
            The variance of all values.

        Raises
        ------
        NonNumericColumnError
            If the data contains non-numerical data.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3, 4, 5])
        >>> column.variance()
        2.5
        """
        if not self.type.is_numeric():
            raise NonNumericColumnError(f"{self.name} is of type {self._type}.")

        return self._data.var()

    # ------------------------------------------------------------------------------------------------------------------
    # Plotting
    # ------------------------------------------------------------------------------------------------------------------

    def plot_boxplot(self) -> Image:
        """
        Plot this column in a boxplot. This function can only plot real numerical data.

        Returns
        -------
        plot: Image
            The plot as an image.

        Raises
        ------
        NonNumericColumnError
            If the data contains non-numerical data.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> boxplot = column.plot_boxplot()
        """
        if not self.type.is_numeric():
            raise NonNumericColumnError(f"{self.name} is of type {self._type}.")

        fig = plt.figure()
        ax = sns.boxplot(data=self._data)
        ax.set(title=self.name)
        ax.set_xticks([])
        plt.tight_layout()

        buffer = io.BytesIO()
        fig.savefig(buffer, format="png")
        plt.close()  # Prevents the figure from being displayed directly
        buffer.seek(0)
        return Image(buffer, ImageFormat.PNG)

    def plot_histogram(self) -> Image:
        """
        Plot a column in a histogram.

        Returns
        -------
        plot: Image
            The plot as an image.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> histogram = column.plot_histogram()
        """
        fig = plt.figure()
        ax = sns.histplot(data=self._data)
        ax.set_xticks(ax.get_xticks())
        ax.set(xlabel=self.name)
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=45,
            horizontalalignment="right",
        )  # rotate the labels of the x Axis to prevent the chance of overlapping of the labels
        plt.tight_layout()

        buffer = io.BytesIO()
        fig.savefig(buffer, format="png")
        plt.close()  # Prevents the figure from being displayed directly
        buffer.seek(0)
        return Image(buffer, ImageFormat.PNG)

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    def to_html(self) -> str:
        r"""
        Return an HTML representation of the column.

        Returns
        -------
        output : str
            The generated HTML.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("test", [1, 2, 3])
        >>> column.to_html()
        '<table border="1" class="dataframe">\n  <thead>\n    <tr style="text-align: right;">\n      <th></th>\n      <th>test</th>\n    </tr>\n  </thead>\n  <tbody>\n    <tr>\n      <th>0</th>\n      <td>1</td>\n    </tr>\n    <tr>\n      <th>1</th>\n      <td>2</td>\n    </tr>\n    <tr>\n      <th>2</th>\n      <td>3</td>\n    </tr>\n  </tbody>\n</table>'
        """
        frame = self._data.to_frame()
        frame.columns = [self.name]

        return frame.to_html(max_rows=self._data.size, max_cols=1)

    # ------------------------------------------------------------------------------------------------------------------
    # IPython integration
    # ------------------------------------------------------------------------------------------------------------------

    def _repr_html_(self) -> str:
        r"""
        Return an HTML representation of the column.

        Returns
        -------
        output : str
            The generated HTML.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("col_1", ['a', 'b', 'c'])
        >>> column._repr_html_()
        '<div>\n<style scoped>\n    .dataframe tbody tr th:only-of-type {\n        vertical-align: middle;\n    }\n\n    .dataframe tbody tr th {\n        vertical-align: top;\n    }\n\n    .dataframe thead th {\n        text-align: right;\n    }\n</style>\n<table border="1" class="dataframe">\n  <thead>\n    <tr style="text-align: right;">\n      <th></th>\n      <th>col_1</th>\n    </tr>\n  </thead>\n  <tbody>\n    <tr>\n      <th>0</th>\n      <td>a</td>\n    </tr>\n    <tr>\n      <th>1</th>\n      <td>b</td>\n    </tr>\n    <tr>\n      <th>2</th>\n      <td>c</td>\n    </tr>\n  </tbody>\n</table>\n</div>'
        """
        frame = self._data.to_frame()
        frame.columns = [self.name]

        return frame.to_html(max_rows=self._data.size, max_cols=1, notebook=True)

    # ------------------------------------------------------------------------------------------------------------------
    # Other
    # ------------------------------------------------------------------------------------------------------------------

    def _count_missing_values(self) -> int:
        """
        Return the number of null values in the column.

        Returns
        -------
        count : int
            The number of null values.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column
        >>> column = Column("col_1", [None, 'a', None])
        >>> column._count_missing_values()
        2
        """
        return self._data.isna().sum()

    # ------------------------------------------------------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------------------------------------------------------

    def _copy(self) -> Column:
        """
        Return a copy of this column.

        Returns
        -------
        column : Column
            The copy of this column.

        """
        return copy.deepcopy(self)



from abc import ABC, abstractmethod

from sklearn.impute import SimpleImputer as sk_SimpleImputer


class ImputerStrategy(ABC):
    """
    The abstract base class of the different imputation strategies supported by the `Imputer`.

    This class is only needed for type annotations. Use the subclasses nested inside `Imputer.Strategy` instead.
    """

    @abstractmethod
    def _augment_imputer(self, imputer: sk_SimpleImputer) -> None:
        """
        Set the imputer strategy of the given imputer.

        Parameters
        ----------
        imputer: SimpleImputer
            The imputer to augment.
        """



from __future__ import annotations

import copy
import io
import warnings
from pathlib import Path
from typing import Any, BinaryIO

import numpy as np
import PIL
from PIL import ImageEnhance, ImageFilter, ImageOps
from PIL.Image import Image as PillowImage
from PIL.Image import open as open_image
from skimage.util import random_noise

from safeds.data.image.typing import ImageFormat
from safeds.exceptions import ClosedBound, OutOfBoundsError


class Image:
    """
    A container for image data.

    Parameters
    ----------
    data : BinaryIO
        The image data as bytes.
    """

    @staticmethod
    def from_jpeg_file(path: str | Path) -> Image:
        """
        Create an image from a JPEG file.

        Parameters
        ----------
        path : str | Path
            The path to the JPEG file.

        Returns
        -------
        image : Image
            The image.
        """
        return Image(
            data=Path(path).open("rb"),
            format_=ImageFormat.JPEG,
        )

    @staticmethod
    def from_png_file(path: str | Path) -> Image:
        """
        Create an image from a PNG file.

        Parameters
        ----------
        path : str | Path
            The path to the PNG file.

        Returns
        -------
        image : Image
            The image.
        """
        return Image(
            data=Path(path).open("rb"),
            format_=ImageFormat.PNG,
        )

    def __init__(self, data: BinaryIO, format_: ImageFormat):
        data.seek(0)

        self._image: PillowImage = open_image(data, formats=[format_.value])
        self._format: ImageFormat = format_

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def format(self) -> ImageFormat:
        """
        Get the image format.

        Returns
        -------
        format : ImageFormat
            The image format.
        """
        return self._format

    @property
    def width(self) -> int:
        """
        Get the width of the image in pixels.

        Returns
        -------
        width : int
            The width of the image.
        """
        return self._image.width

    @property
    def height(self) -> int:
        """
        Get the height of the image in pixels.

        Returns
        -------
        height : int
            The height of the image.
        """
        return self._image.height

    # ------------------------------------------------------------------------------------------------------------------
    # Conversion
    # ------------------------------------------------------------------------------------------------------------------

    def to_jpeg_file(self, path: str | Path) -> None:
        """
        Save the image as a JPEG file.

        Parameters
        ----------
        path : str | Path
            The path to the JPEG file.
        """
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        self._image.save(path, format="jpeg")

    def to_png_file(self, path: str | Path) -> None:
        """
        Save the image as a PNG file.

        Parameters
        ----------
        path : str | Path
            The path to the PNG file.
        """
        Path(path).parent.mkdir(parents=True, exist_ok=True)
        self._image.save(path, format="png")

    # ------------------------------------------------------------------------------------------------------------------
    # IPython integration
    # ------------------------------------------------------------------------------------------------------------------

    def __eq__(self, other: Any) -> bool:
        """
        Compare two images.

        Parameters
        ----------
        other: The image to compare to.

        Returns
        -------
        equals : bool
            Whether the two images contain equal pixel data.

        """
        if not isinstance(other, Image):
            return NotImplemented
        return self._image.tobytes() == other._image.tobytes()

    def _repr_jpeg_(self) -> bytes | None:
        """
        Return a JPEG image as bytes.

        If the image is not a JPEG, return None.

        Returns
        -------
        jpeg : bytes | None
            The image as JPEG.
        """
        if self._format != ImageFormat.JPEG:
            return None

        buffer = io.BytesIO()
        self._image.save(buffer, format="jpeg")
        buffer.seek(0)
        return buffer.read()

    def _repr_png_(self) -> bytes | None:
        """
        Return a PNG image as bytes.

        If the image is not a PNG, return None.

        Returns
        -------
        png : bytes | None
            The image as PNG.
        """
        if self._format != ImageFormat.PNG:
            return None

        buffer = io.BytesIO()
        self._image.save(buffer, format="png")
        buffer.seek(0)
        return buffer.read()

    # ------------------------------------------------------------------------------------------------------------------
    # Transformations
    # ------------------------------------------------------------------------------------------------------------------

    def resize(self, new_width: int, new_height: int) -> Image:
        """
        Return a new `Image` that has been resized to a given size.

        The original image is not modified.

        Returns
        -------
        result : Image
            The image with the given width and height.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = image_copy._image.resize((new_width, new_height))
        return image_copy

    def convert_to_grayscale(self) -> Image:
        """
        Return a new `Image` that is converted to grayscale.

        The original image is not modified.

        Returns
        -------
        result : Image
            The grayscale image.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = image_copy._image.convert("L")
        return image_copy

    def crop(self, x: int, y: int, width: int, height: int) -> Image:
        """
        Return a new `Image` that has been cropped to a given bounding rectangle.

        The original image is not modified.

        Parameters
        ----------
        x: the x coordinate of the top-left corner of the bounding rectangle
        y: the y coordinate of the top-left corner of the bounding rectangle
        width:  the width of the bounding rectangle
        height:  the height of the bounding rectangle

        Returns
        -------
        result : Image
            The cropped image.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = image_copy._image.crop((x, y, (x + width), (y + height)))
        return image_copy

    def flip_vertically(self) -> Image:
        """
        Return a new `Image` that is flipped vertically (horizontal axis, flips up-down and vice versa).

        The original image is not modified.

        Returns
        -------
        result : Image
            The flipped image.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = self._image.transpose(PIL.Image.FLIP_TOP_BOTTOM)
        return image_copy

    def flip_horizontally(self) -> Image:
        """
        Return a new `Ìmage` that is flipped horizontally (vertical axis, flips left-right and vice versa).

        The original image is not modified.

        Returns
        -------
        result : Image
            The flipped image.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = self._image.transpose(PIL.Image.FLIP_LEFT_RIGHT)
        return image_copy

    def adjust_brightness(self, factor: float) -> Image:
        """
        Return a new `Image` with an adjusted brightness.

        The original image is not modified.

        Parameters
        ----------
        factor: float
            The brightness factor.
            1.0 will not change the brightness.
            Below 1.0 will result in a darker image.
            Above 1.0 will resolut in a brighter image.
            Has to be bigger than or equal to 0 (black).

        Returns
        -------
        result: Image
            The Image with adjusted brightness.
        """
        if factor < 0:
            raise OutOfBoundsError(factor, name="factor", lower_bound=ClosedBound(0))
        elif factor == 1:
            warnings.warn(
                "Brightness adjustment factor is 1.0, this will not make changes to the image.",
                UserWarning,
                stacklevel=2,
            )

        image_copy = copy.deepcopy(self)
        image_copy._image = ImageEnhance.Brightness(image_copy._image).enhance(factor)
        return image_copy

    def add_gaussian_noise(self, standard_deviation: float) -> Image:
        """
        Return a new `Image` with Gaussian noise added to the image.

        The original image is not modified.

        Parameters
        ----------
        standard_deviation : float
            The standard deviation of the Gaussian distribution. Has to be bigger than or equal to 0.

        Returns
        -------
        result : Image
            The image with added Gaussian noise.

        Raises
        ------
        OutOfBoundsError
            If standard_deviation is smaller than 0.
        """
        if standard_deviation < 0:
            raise OutOfBoundsError(standard_deviation, name="standard_deviation", lower_bound=ClosedBound(0))

        # noinspection PyTypeChecker
        image_as_array = np.asarray(self._image)
        noisy_image_as_array = random_noise(
            image_as_array,
            mode="gaussian",
            var=standard_deviation**2,
            rng=42,
            clip=True,
        )
        noisy_image = PIL.Image.fromarray(np.uint8(255 * noisy_image_as_array))

        image_copy = copy.deepcopy(self)
        image_copy._image = noisy_image
        return image_copy

    def adjust_contrast(self, factor: float) -> Image:
        """
        Return a new `Image` with adjusted contrast.

        The original image is not modified.

        Parameters
        ----------
        factor: float
            If factor > 1, increase contrast of image.
            If factor = 1, no changes will be made.
            If factor < 1, make image greyer.
            Has to be bigger than or equal to 0 (gray).

        Returns
        -------
        image: Image
            New image with adjusted contrast.
        """
        if factor < 0:
            raise OutOfBoundsError(factor, name="factor", lower_bound=ClosedBound(0))
        elif factor == 1:
            warnings.warn(
                "Contrast adjustment factor is 1.0, this will not make changes to the image.",
                UserWarning,
                stacklevel=2,
            )

        image_copy = copy.deepcopy(self)
        image_copy._image = ImageEnhance.Contrast(image_copy._image).enhance(factor)
        return image_copy

    def adjust_color_balance(self, factor: float) -> Image:
        """
        Return a new `Image` with adjusted color balance.

        The original image is not modified.

        Parameters
        ----------
        factor: float
            If factor > 1, increase color balance of image.
            If factor = 1, no changes will be made.
            If factor < 1, make image greyer.
            Has to be bigger than or equal to 0.

        Returns
        -------
        image: Image
            The new, adjusted image.
        """
        if factor < 0:
            raise OutOfBoundsError(factor, name="factor", lower_bound=ClosedBound(0))
        elif factor == 1:
            warnings.warn(
                "Color adjustment factor is 1.0, this will not make changes to the image.",
                UserWarning,
                stacklevel=2,
            )

        image_copy = copy.deepcopy(self)
        image_copy._image = ImageEnhance.Color(image_copy._image).enhance(factor)
        return image_copy

    def blur(self, radius: int) -> Image:
        """
        Return a blurred version of the image.

        The original image is not modified.

        Parameters
        ----------
        radius : int
             Radius is directly proportional to the blur value. The radius is equal to the amount of pixels united in
             each direction. A radius of 1 will result in a united box of 9 pixels.

        Returns
        -------
        result : Image
            The blurred image.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = image_copy._image.filter(ImageFilter.BoxBlur(radius))
        return image_copy

    def sharpen(self, factor: float) -> Image:
        """
        Return a sharpened version of the image.

        The original image is not modified.

        Parameters
        ----------
        factor : float
            The amount of sharpness to be applied to the image. Factor 1.0 is considered to be neutral and does not make
            any changes.

        Returns
        -------
        result : Image
            The image sharpened by the given factor.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = ImageEnhance.Sharpness(image_copy._image).enhance(factor)
        return image_copy

    def invert_colors(self) -> Image:
        """
        Return a new image with colors inverted.

        The original image is not modified.

        Returns
        -------
        result : Image
            The image with inverted colors.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = ImageOps.invert(image_copy._image.convert("RGB"))
        return image_copy

    def rotate_right(self) -> Image:
        """
        Return a new `Image` that is rotated 90 degrees clockwise.

        The original image is not modified.

        Returns
        -------
        result : Image
            The image rotated 90 degrees clockwise.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = image_copy._image.rotate(270, expand=True)
        return image_copy

    def rotate_left(self) -> Image:
        """
        Return a new `Image` that is rotated 90 degrees counter-clockwise.

        The original image is not modified.

        Returns
        -------
        result : Image
            The image rotated 90 degrees counter-clockwise.
        """
        image_copy = copy.deepcopy(self)
        image_copy._image = image_copy._image.rotate(90, expand=True)
        return image_copy

    def find_edges(self) -> Image:
        """
        Return a grayscale version of the image with the edges highlighted.

        The original image is not modified.

        Returns
        -------
        result : Image
            The image with edges found.
        """
        image_copy = copy.deepcopy(self)
        image_copy = image_copy.convert_to_grayscale()
        image_copy._image = image_copy._image.filter(ImageFilter.FIND_EDGES)
        return image_copy



"""Classes for transforming tabular data."""

from ._discretizer import Discretizer
from ._imputer import Imputer
from ._label_encoder import LabelEncoder
from ._one_hot_encoder import OneHotEncoder
from ._range_scaler import RangeScaler
from ._standard_scaler import StandardScaler
from ._table_transformer import InvertibleTableTransformer, TableTransformer

__all__ = [
    "Discretizer",
    "Imputer",
    "InvertibleTableTransformer",
    "LabelEncoder",
    "OneHotEncoder",
    "RangeScaler",
    "StandardScaler",
    "TableTransformer",
]
